
import os
import numpy as np
import torch
import pytorch_lightning as pl

from sklearn.base import BaseEstimator, ClassifierMixin
from torch.utils.data import DataLoader
from pytorch_lightning.callbacks import ModelCheckpoint

from neuralnets.data.datasets import LabeledVolumeDataset
from neuralnets.networks.unet import UNet2D, UNet3D
from neuralnets.util.losses import CrossEntropyLoss


def _correct_type(param, values):

    vs = values.split(';')
    values = []
    for v in vs:
        if param == 'feature_maps' or param == 'levels' or param == 'epochs':
            v_ = int(v)
        elif param == 'skip_connections' or param == 'residual_connections':
            v_ = bool(int(v))
        elif param == 'dropout' or param == 'lr':
            v_ = float(v)
        elif param == 'input_shape':
            v_ = [int(item) for item in v.split(',')]
        else:
            v_ = v
        values.append(v_)

    return param, values


def parse_search_grid(sg_str):

    sg = sg_str.split('#')
    search_grid = {}
    for s in sg:
        param, values = s.split(':')
        param, values = _correct_type(param, values)
        search_grid[param] = values

    return search_grid


class PLClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self, epochs=10, gpus=(0,), accelerator='dp', log_dir='logs', log_freq=50, log_refresh_rate=None,
                 train_batch_size=1, test_batch_size=1, num_workers=1, device=0, orientations=(0,),
                 normalization='unit', transform=None):

        # trainer parameters
        self.epochs = epochs
        self.gpus = gpus
        self.accelerator = accelerator
        self.log_dir = log_dir
        self.log_freq = log_freq
        self.log_refresh_rate = log_refresh_rate

        # inference parameters
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.num_workers = num_workers
        self.device = device  # temporary while multi GPU inference is not supported
        self.orientations = orientations
        self.normalization = normalization
        self.transform = transform

        # initialize the trainer
        checkpoint_callback = ModelCheckpoint(save_top_k=0)
        self.callbacks = [checkpoint_callback]

        # initialize the model and trainer: should be set in an implementation!
        self.trainer = None
        self.model = None


    def fit(self, X, y):

        # construct dataloader
        train = LabeledVolumeDataset(X, y, input_shape=self.model.input_shape, batch_size=self.train_batch_size,
                                     transform=self.transform)
        loader = DataLoader(train, batch_size=self.train_batch_size, num_workers=self.num_workers, pin_memory=True)

        # train the network
        self.trainer.fit(self.model, loader)

        return self


    def predict(self, X, y):

        segmentation = self.model.segment(X, self.model, self.model.input_shape, in_channels=self.model.in_channels,
                                          batch_size=self.test_batch_size, device=self.device,
                                          orientations=self.orientations, normalization=self.normalization)

        return segmentation


    def score(self, X, y, sample_weight=None):

        # validate each model state and save the metrics
        checkpoints = os.listdir(self.trainer.checkpoint_callback.dirpath)
        checkpoints.sort()
        metrics = np.zeros(len(checkpoints))
        for i, ckpt in enumerate(checkpoints):
            ckpt_path = os.path.join(self.trainer.checkpoint_callback.dirpath, ckpt)
            self.model.load_state_dict(torch.load(ckpt_path))
            metrics[i] = self.model.validate(X, y, self.model.input_shape, in_channels=self.model.in_channels,
                                             classes_of_interest=self.model.coi, batch_size=self.test_batch_size,
                                             device=self.device, orientations=self.orientations,
                                             normalization=self.normalization, report=False)

        # find the best model state
        j = np.argmax(metrics)
        metric = metrics[j]

        # remove the remaining checkpoints
        for i, ckpt in enumerate(checkpoints):
            if i != j:
                ckpt_path = os.path.join(self.trainer.checkpoint_callback.dirpath, ckpt)
                os.remove(ckpt_path)

        return metric


class UNet2DClassifier(PLClassifier):

    def __init__(self, epochs=10, gpus=(0,), accelerator='dp', log_dir='logs', log_freq=50, log_refresh_rate=None,
                 train_batch_size=1, test_batch_size=1, num_workers=1, device=0, orientations=(0,),
                 normalization='unit', transform=None, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1),
                 feature_maps=64, levels=4, skip_connections=True, residual_connections=False, norm='instance',
                 activation='relu', dropout=0.0, loss_fn=CrossEntropyLoss(), lr=1e-3):
        super().__init__(epochs=epochs, gpus=gpus, accelerator=accelerator, log_dir=log_dir, log_freq=log_freq,
                         log_refresh_rate=log_refresh_rate, train_batch_size=train_batch_size,
                         test_batch_size=test_batch_size, num_workers=num_workers, device=device,
                         orientations=orientations, normalization=normalization, transform=transform)

        # parameters
        self.input_shape = input_shape
        self.in_channels = in_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.skip_connections = skip_connections
        self.residual_connections = residual_connections
        self.dropout = dropout
        self.norm = norm
        self.activation = activation
        self.coi = coi
        self.loss_fn = loss_fn
        self.lr = lr


    def fit(self, X, y):

        # initialize model and trainer
        self.model = UNet2D(input_shape=self.input_shape, in_channels=self.in_channels, feature_maps=self.feature_maps,
                            levels=self.levels, skip_connections=self.skip_connections,
                            residual_connections=self.residual_connections, dropout_enc=self.dropout,
                            dropout_dec=self.dropout, norm=self.norm, activation=self.activation, coi=self.coi,
                            loss_fn=self.loss_fn, lr=self.lr)
        self.trainer = pl.Trainer(max_epochs=int(self.epochs), gpus=self.gpus, accelerator=self.accelerator,
                                  default_root_dir=self.log_dir, flush_logs_every_n_steps=self.log_freq,
                                  log_every_n_steps=self.log_freq, callbacks=self.callbacks,
                                  progress_bar_refresh_rate=self.log_refresh_rate)

        # construct dataloader
        train = LabeledVolumeDataset(X, y, input_shape=self.model.input_shape, in_channels=self.in_channels,
                                     batch_size=self.train_batch_size, transform=self.transform)
        loader = DataLoader(train, batch_size=self.train_batch_size, num_workers=self.num_workers, pin_memory=True)

        # train the network
        self.trainer.fit(self.model, loader)

        return self


class UNet3DClassifier(PLClassifier):

    def __init__(self, epochs=10, gpus=(0,), accelerator='dp', log_dir='logs', log_freq=50, log_refresh_rate=None,
                 train_batch_size=1, test_batch_size=1, num_workers=1, device=0, orientations=(0,),
                 normalization='unit', transform=None, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1),
                 feature_maps=64, levels=4, skip_connections=True, residual_connections=False, norm='instance',
                 activation='relu', dropout=0.0, loss_fn=CrossEntropyLoss(), lr=1e-3):
        super().__init__(epochs=epochs, gpus=gpus, accelerator=accelerator, log_dir=log_dir, log_freq=log_freq,
                         log_refresh_rate=log_refresh_rate, train_batch_size=train_batch_size,
                         test_batch_size=test_batch_size, num_workers=num_workers, device=device,
                         orientations=orientations, normalization=normalization, transform=transform)

        # parameters
        self.input_shape = input_shape
        self.in_channels = in_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.skip_connections = skip_connections
        self.residual_connections = residual_connections
        self.dropout = dropout
        self.norm = norm
        self.activation = activation
        self.coi = coi
        self.loss_fn = loss_fn
        self.lr = lr


    def fit(self, X, y):

        # initialize model and trainer
        self.model = UNet3D(input_shape=self.input_shape, in_channels=self.in_channels, feature_maps=self.feature_maps,
                            levels=self.levels, skip_connections=self.skip_connections,
                            residual_connections=self.residual_connections, dropout_enc=self.dropout,
                            dropout_dec=self.dropout, norm=self.norm, activation=self.activation, coi=self.coi,
                            loss_fn=self.loss_fn, lr=self.lr)
        self.trainer = pl.Trainer(max_epochs=int(self.epochs), gpus=self.gpus, accelerator=self.accelerator,
                                  default_root_dir=self.log_dir, flush_logs_every_n_steps=self.log_freq,
                                  log_every_n_steps=self.log_freq, callbacks=self.callbacks,
                                  progress_bar_refresh_rate=self.log_refresh_rate)

        # construct dataloader
        train = LabeledVolumeDataset(X, y, input_shape=self.model.input_shape, in_channels=self.in_channels,
                                     batch_size=self.train_batch_size, transform=self.transform)
        loader = DataLoader(train, batch_size=self.train_batch_size, num_workers=self.num_workers, pin_memory=True)

        # train the network
        self.trainer.fit(self.model, loader)

        return self




import numpy as np
from scipy.spatial.distance import directed_hausdorff


def _iou_part_metrics(y_true, y_pred, w=None, max_elements=1000000):
    """
    computes the partial iou metrics in a memory efficient way

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :param max_elements: maximum number of elements for block based computing
    :return: intersection, true_total, pred_total
    """

    # flatten, apply masking if necessary and binarize
    y_true = y_true.flatten()
    y_pred = y_pred.flatten()
    if w is not None:
        w = w.flatten()
        y_true = y_true[w]
        y_pred = y_pred[w]
    y_true = y_true > 0.5
    y_pred = y_pred > 0.5

    # compute metrics
    intersection = 0
    true_total = 0
    pred_total = 0
    n_blocks = y_true.size // max_elements
    for n in range(n_blocks + 1):
        # find start and stopping point
        start = max_elements * n
        stop = np.minimum(max_elements * (n+1), y_true.size)

        # block to process
        y_true_block = y_true[start:stop]
        y_pred_block = y_pred[start:stop]

        # processing
        intersection += np.sum(y_true_block * y_pred_block)
        true_total += np.sum(y_true_block)
        pred_total += np.sum(y_pred_block)

    return intersection, true_total, pred_total


def iou(y_true, y_pred, w=None):
    """
    Intersection-over-union metric

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the Jaccard index
    """

    # compute jaccard score
    intersection, true_total, pred_total = _iou_part_metrics(y_true, y_pred, w=w)
    union = true_total + pred_total - intersection

    return (intersection + 1) / (union + 1)


def dice(y_true, y_pred, w=None):
    """
    Dice coefficient between two segmentations

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the Jaccard index
    """

    j = iou(y_true, y_pred, w=w)

    return 2 * j / (1 + j)


def _conf_matrix_metrics(y_true, y_pred, w=None, max_elements=1000000):
    """
    computes the confusion matrix metrics in a memory efficient way

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :param max_elements: maximum number of elements for block based computing
    :return: the Jaccard index, accuracy, balanced accuracy, precision, recall and f1-score
    """

    # flatten, apply masking if necessary and binarize
    y_true = y_true.flatten()
    y_pred = y_pred.flatten()
    if w is not None:
        w = w.flatten()
        y_true = y_true[w]
        y_pred = y_pred[w]
    y_true = y_true > 0.5
    y_pred = y_pred > 0.5

    # compute metrics
    tp = 0
    tn = 0
    fp = 0
    fn = 0
    n_blocks = y_true.size // max_elements
    for n in range(n_blocks):
        # find start and stopping point
        start = max_elements * n
        stop = np.minimum(max_elements * (n+1), y_true.size)

        # block to process
        y_true_block = y_true[start:stop]
        y_pred_block = y_pred[start:stop]

        # processing
        tp += np.sum(y_true_block * y_pred_block)
        tn += np.sum((1 - y_true_block) * (1 - y_pred_block))
        fp += np.sum((1 - y_true_block) * y_pred_block)
        fn += np.sum(y_true_block * (1 - y_pred_block))

    return tp, tn, fp, fn


def accuracy_metrics(y_true, y_pred, w=None):
    """
    Accuracy metrics between two segmentations (accuracy, balanced accuracy, precision, recall and f1-score)

    :param y_true: (N1, N2, ...) array of the true labels
    :param y_pred: (N1, N2, ...) array of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking array
    :return: the accuracy, balanced accuracy, precision, recall and f1-score
    """

    # compute accuracy metrics
    tp, tn, fp, fn = _conf_matrix_metrics(y_true, y_pred, w=w)
    total = tp + tn + fp + fn
    accuracy = (tp + tn + 1) / (total + 1)
    recall = (tp + 1) / (tp + fn + 1)
    specificity = (tn + 1) / (tn + fp + 1)
    balanced_accuracy = (recall + specificity) / 2
    precision = (tp + 1) / (tp + fp + 1)
    f1 = 2 * (precision * recall) / (precision + recall)

    return accuracy, balanced_accuracy, precision, recall, f1


def hausdorff_distance(x, y):
    """
    Hausdorff distance between two segmentations

    :param x: array
    :param y: array
    :return: the hausdorff distance
    """

    # binarize
    x = x > 0.5
    y = y > 0.5

    hd_0 = 0
    hd_1 = 0
    hd = 0
    for i in range(x.shape[0]):
        hd_0 += directed_hausdorff(x[i, ...], y[i, ...])[0]
        hd_1 += directed_hausdorff(y[i, ...], x[i, ...])[0]
        hd += max(hd_0, hd_1)
    hd_0 /= x.shape[0]
    hd_1 /= x.shape[0]
    hd /= x.shape[0]

    return hd, hd_0, hd_1



"""
    This is a script that illustrates training a 2D U-Net
"""

"""
    Necessary libraries
"""
import argparse
import yaml

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from neuralnets.data.datasets import LabeledVolumeDataset, LabeledSlidingWindowDataset
from neuralnets.networks.unet import UNet2D
from neuralnets.util.augmentation import *
from neuralnets.util.io import print_frm
from neuralnets.util.tools import set_seed, parse_params

from multiprocessing import freeze_support

if __name__ == '__main__':
    freeze_support()

    """
        Parse all the arguments
    """
    print_frm('Parsing arguments')
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", "-c", help="Path to the configuration file", type=str, default='unet_2d.yaml')
    args = parser.parse_args()
    with open(args.config) as file:
        params = parse_params(yaml.load(file, Loader=yaml.FullLoader))

    """
    Fix seed (for reproducibility)
    """
    set_seed(params['seed'])

    """
        Load the data
    """
    print_frm('Loading data')
    input_shape = (1, *(params['input_size']))
    split = params['train_val_test_split']
    transform = Compose([Rotate90(), Flip(prob=0.5, dim=0), Flip(prob=0.5, dim=1), ContrastAdjust(adj=0.1),
                         RandomDeformation(), AddNoise(sigma_max=0.05), CleanDeformedLabels(params['coi'])])
    train = LabeledVolumeDataset(params['data'], params['labels'], input_shape=input_shape,
                                 in_channels=params['in_channels'], type=params['type'],
                                 batch_size=params['train_batch_size'], transform=transform, range_split=(0, split[0]),
                                 range_dir=params['split_orientation'])
    val = LabeledSlidingWindowDataset(params['data'], params['labels'], input_shape=input_shape,
                                      in_channels=params['in_channels'], type=params['type'],
                                      batch_size=params['test_batch_size'], range_split=(split[0], split[1]),
                                      range_dir=params['split_orientation'])
    test = LabeledSlidingWindowDataset(params['data'], params['labels'], input_shape=input_shape,
                                       in_channels=params['in_channels'], type=params['type'],
                                       batch_size=params['test_batch_size'], range_split=(split[1], 1),
                                       range_dir=params['split_orientation'])
    train_loader = DataLoader(train, batch_size=params['train_batch_size'], num_workers=params['num_workers'],
                              pin_memory=True)
    val_loader = DataLoader(val, batch_size=params['test_batch_size'], num_workers=params['num_workers'],
                            pin_memory=True)
    test_loader = DataLoader(test, batch_size=params['test_batch_size'], num_workers=params['num_workers'],
                             pin_memory=True)

    """
        Build the network
    """
    print_frm('Building the network')
    net = UNet2D(in_channels=params['in_channels'], feature_maps=params['fm'], levels=params['levels'],
                 dropout_enc=params['dropout'], dropout_dec=params['dropout'], norm=params['norm'],
                 activation=params['activation'], coi=params['coi'], loss_fn=params['loss'])

    """
        Train the network
    """
    print_frm('Starting training')
    print_frm('Training with loss: %s' % params['loss'])
    lr_monitor = pl.callbacks.LearningRateMonitor(logging_interval='step')
    trainer = pl.Trainer(max_epochs=params['epochs'], gpus=params['gpus'], accelerator=params['accelerator'],
                         default_root_dir=params['log_dir'], flush_logs_every_n_steps=params['log_freq'],
                         log_every_n_steps=params['log_freq'], callbacks=[lr_monitor],
                         progress_bar_refresh_rate=params['log_refresh_rate'])
    trainer.fit(net, train_loader, val_loader)

    """
        Testing the network
    """
    print_frm('Testing network')
    trainer.test(net, test_loader)



from torch.optim.lr_scheduler import ReduceLROnPlateau
import pytorch_lightning as pl
from pytorch_lightning.metrics.functional import iou
import os

from neuralnets.networks.blocks import *
from neuralnets.util.torch.metrics import iou
from neuralnets.util.tools import *
from neuralnets.util.augmentation import *
from neuralnets.util.losses import get_loss_function
from neuralnets.util.validation import segment, validate
from neuralnets.util.visualization import overlay, COLORS


class UNetEncoder(nn.Module):
    """
    U-Net encoder base class

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional dense_blocks: specify use of dense blocks
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu',
                 dense_blocks=False):
        super().__init__()

        self.features = nn.Sequential()
        self.in_channels = in_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.norm = norm
        self.dropout = dropout
        self.activation = activation
        self.dense_blocks = dense_blocks


class UNetDecoder(nn.Module):
    """
    U-Net decoder base class

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional residual_connections: use residual connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, residual_connections=False,
                 norm='instance', dropout=0.0, activation='relu'):
        super().__init__()

        self.features = nn.Sequential()
        self.out_channels = out_channels
        self.feature_maps = feature_maps
        self.levels = levels
        self.skip_connections = skip_connections
        self.residual_connections = residual_connections
        self.norm = norm
        self.dropout = dropout
        self.activation = activation


class UNetEncoder2D(UNetEncoder):
    """
    2D U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu'):
        super().__init__(in_channels=in_channels, feature_maps=feature_maps, levels=levels, norm=norm, dropout=dropout,
                         activation=activation)

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # convolutional block
            conv_block = UNetConvBlock2D(in_features, out_features, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)


            # pooling
            pool = nn.MaxPool2d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = UNetConvBlock2D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, norm=norm,
                                           dropout=dropout, activation=activation)

    def forward(self, inputs):
        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class UNetEncoder3D(UNetEncoder):
    """
    3D U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu'):
        super().__init__(in_channels=in_channels, feature_maps=feature_maps, levels=levels, norm=norm, dropout=dropout,
                         activation=activation)

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # convolutional block
            conv_block = UNetConvBlock3D(in_features, out_features, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

            # pooling
            pool = nn.MaxPool3d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = UNetConvBlock3D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, norm=norm,
                                           dropout=dropout, activation=activation)

    def forward(self, inputs):
        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class UNetDecoder2D(UNetDecoder):
    """
    2D U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional residual_connections: use residual connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, residual_connections=False,
                 norm='instance', dropout=0.0, activation='relu'):
        super().__init__(out_channels=out_channels, feature_maps=feature_maps, levels=levels,
                         skip_connections=skip_connections, residual_connections=residual_connections, norm=norm,
                         dropout=dropout, activation=activation)

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections and not residual_connections:
                conv_block = UNetConvBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = UNetConvBlock2D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv2d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        outputs = inputs
        for i in range(self.levels):
            if self.residual_connections:
                outputs = encoder_outputs[self.levels - i - 1] + \
                          getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # residual connection
            elif self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[self.levels - i - 1],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class UNetDecoder3D(UNetDecoder):
    """
    3D U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional residual_connections: use residual connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, residual_connections=False,
                 norm='instance', dropout=0.0, activation='relu'):
        super().__init__(out_channels=out_channels, feature_maps=feature_maps, levels=levels,
                         skip_connections=skip_connections, residual_connections=residual_connections, norm=norm,
                         dropout=dropout, activation=activation)

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections and not residual_connections:
                conv_block = UNetConvBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = UNetConvBlock3D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                             norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv3d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        outputs = inputs
        for i in range(self.levels):
            if self.residual_connections:
                outputs = encoder_outputs[self.levels - i - 1] + \
                          getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # residual connection
            elif self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[self.levels - i - 1],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class DenseUNetEncoder2D(UNetEncoder):
    """
    2D Dense U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional num_layers: number of dense layers
    :param optional k: how many filters to add each layer
    :param optional bn_size: multiplicative factor for number of bottle neck layers
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu',
                 num_layers=4, k=16, bn_size=2):
        super().__init__(in_channels=in_channels, feature_maps=feature_maps, levels=levels, norm=norm, dropout=dropout,
                         activation=activation)

        self.num_layers = num_layers
        self.k = k
        self.bn_size = bn_size

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # dense convolutional block
            conv_block = DenseBlock2D(in_features, out_features, num_layers, k, bn_size, norm=norm, dropout=dropout,
                                      activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

            # pooling
            pool = nn.MaxPool2d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = DenseBlock2D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, num_layers, k,
                                        bn_size, norm=norm, dropout=dropout, activation=activation)

    def forward(self, inputs):
        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class DenseUNetEncoder3D(UNetEncoder):
    """
    3D Dense U-Net encoder

    :param optional in_channels: number of input channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional num_layers: number of dense layers
    :param optional k: how many filters to add each layer
    :param optional bn_size: multiplicative factor for number of bottle neck layers
    """

    def __init__(self, in_channels=1, feature_maps=64, levels=4, norm='instance', dropout=0.0, activation='relu',
                 num_layers=4, k=16, bn_size=2):
        super().__init__(in_channels=in_channels, feature_maps=feature_maps, levels=levels, norm=norm, dropout=dropout,
                         activation=activation)

        self.num_layers = num_layers
        self.k = k
        self.bn_size = bn_size

        in_features = in_channels
        for i in range(levels):
            out_features = (2 ** i) * feature_maps

            # dense convolutional block
            conv_block = DenseBlock3D(in_features, out_features, num_layers, k, bn_size, norm=norm, dropout=dropout,
                                      activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

            # pooling
            pool = nn.MaxPool3d(kernel_size=2, stride=2)
            self.features.add_module('pool%d' % (i + 1), pool)

            # input features for next block
            in_features = out_features

        # center (lowest) block
        self.center_conv = DenseBlock3D(2 ** (levels - 1) * feature_maps, 2 ** levels * feature_maps, num_layers, k,
                                        bn_size, norm=norm, dropout=dropout, activation=activation)

    def forward(self, inputs):
        encoder_outputs = []  # for decoder skip connections

        outputs = inputs
        for i in range(self.levels):
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            encoder_outputs.append(outputs)
            outputs = getattr(self.features, 'pool%d' % (i + 1))(outputs)

        outputs = self.center_conv(outputs)

        return encoder_outputs, outputs


class DenseUNetDecoder2D(UNetDecoder):
    """
    2D Dense U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional residual_connections: use residual connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional num_layers: number of dense layers
    :param optional k: how many filters to add each layer
    :param optional bn_size: multiplicative factor for number of bottle neck layers
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, residual_connections=False,
                 norm='instance', dropout=0.0, activation='relu', num_layers=4, k=16, bn_size=2):
        super().__init__(out_channels=out_channels, feature_maps=feature_maps, levels=levels,
                         skip_connections=skip_connections, residual_connections=residual_connections, norm=norm,
                         dropout=dropout, activation=activation)

        self.num_layers = num_layers
        self.k = k
        self.bn_size = bn_size

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections and not residual_connections:
                conv_block = DenseBlock2D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                          num_layers, k, bn_size, norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = DenseBlock2D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                          num_layers, k, bn_size, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv2d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        outputs = inputs
        for i in range(self.levels):
            if self.residual_connections:
                outputs = encoder_outputs[self.levels - i - 1] + \
                          getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # residual connection
            elif self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[self.levels - i - 1],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class DenseUNetDecoder3D(UNetDecoder):
    """
    3D Dense U-Net decoder

    :param optional out_channels: number of output channels
    :param optional feature_maps: number of initial feature maps
    :param optional levels: levels of the encoder
    :param optional skip_connections: use skip connections or not
    :param optional residual_connections: use residual connections or not
    :param optional norm: specify normalization ("batch", "instance" or None)
    :param optional dropout: dropout factor
    :param optional activation: specify activation function ("relu", "sigmoid" or None)
    :param optional num_layers: number of dense layers
    :param optional k: how many filters to add each layer
    :param optional bn_size: multiplicative factor for number of bottle neck layers
    """

    def __init__(self, out_channels=2, feature_maps=64, levels=4, skip_connections=True, residual_connections=False,
                 norm='instance', dropout=0.0, activation='relu', num_layers=4, k=16, bn_size=2):
        super().__init__(out_channels=out_channels, feature_maps=feature_maps, levels=levels,
                         skip_connections=skip_connections, residual_connections=residual_connections, norm=norm,
                         dropout=dropout, activation=activation)

        self.num_layers = num_layers
        self.k = k
        self.bn_size = bn_size

        for i in range(levels):

            # upsampling block
            upconv = UNetUpSamplingBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                           deconv=True)
            self.features.add_module('upconv%d' % (i + 1), upconv)

            # convolutional block
            if skip_connections and not residual_connections:
                conv_block = DenseBlock3D(2 ** (levels - i) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                          num_layers, k, bn_size, norm=norm, dropout=dropout, activation=activation)
            else:
                conv_block = DenseBlock3D(2 ** (levels - i - 1) * feature_maps, 2 ** (levels - i - 1) * feature_maps,
                                          num_layers, k, bn_size, norm=norm, dropout=dropout, activation=activation)
            self.features.add_module('convblock%d' % (i + 1), conv_block)

        # output layer
        self.output = nn.Conv3d(feature_maps, out_channels, kernel_size=1)

    def forward(self, inputs, encoder_outputs):

        decoder_outputs = []

        outputs = inputs
        for i in range(self.levels):
            if self.residual_connections:
                outputs = encoder_outputs[self.levels - i - 1] + \
                          getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # residual connection
            elif self.skip_connections:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(encoder_outputs[self.levels - i - 1],
                                                                       outputs)  # also deals with concat
            else:
                outputs = getattr(self.features, 'upconv%d' % (i + 1))(outputs)  # no concat
            outputs = getattr(self.features, 'convblock%d' % (i + 1))(outputs)
            decoder_outputs.append(outputs)

        outputs = self.output(outputs)

        return decoder_outputs, outputs


class UNet(pl.LightningModule):

    def __init__(self, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1), feature_maps=64, levels=4,
                 skip_connections=True, residual_connections=False, norm='instance', activation='relu', dropout_enc=0.0,
                 dropout_dec=0.0, loss_fn="ce", lr=1e-3):
        super().__init__()

        # parameters
        if isinstance(input_shape, str):
            self.input_shape = [int(item) for item in input_shape.split(',')]
        else:  # assuming list-like object
            self.input_shape = input_shape
        self.in_channels = int(in_channels)
        self.c = self.in_channels // 2
        self.coi = coi
        self.out_channels = len(coi)
        self.feature_maps = int(feature_maps)
        self.levels = int(levels)
        self.skip_connections = bool(skip_connections)
        self.residual_connections = bool(residual_connections)
        self.norm = norm
        self.dropout_enc = float(dropout_enc)
        self.dropout_dec = float(dropout_dec)
        self.activation = activation
        self.loss_fn = get_loss_function(loss_fn)
        self.lr = float(lr)

        self.train_batch_id = 0
        self.val_batch_id = 0

    def forward(self, x):

        # contractive path
        encoder_outputs, final_output = self.encoder(x)

        # expansive path
        decoder_outputs, outputs = self.decoder(final_output, encoder_outputs)

        return outputs

    def training_step(self, batch, batch_idx):

        # transfer to suitable device and get labels
        x, y = batch

        # forward prop
        y_pred = self(x)

        # compute loss
        loss = self.loss_fn(y_pred, y[:, 0, ...])

        # compute iou
        mIoU = 0
        y_pred = torch.softmax(y_pred, dim=1)
        for c in range(y_pred.size(1)):
            mIoU += iou(y_pred[:, c:c+1, ...], y == c, w=(y != 255))
        mIoU /= y_pred.size(1)
        self.log('train/mIoU', mIoU, prog_bar=True)
        self.log('train/loss', loss)

        # log images
        if batch_idx == self.train_batch_id:
            self._log_predictions(x, y, y_pred, prefix='train')

        return loss

    def validation_step(self, batch, batch_idx):

        # transfer to suitable device and get labels
        x, y = batch

        # forward prop
        y_pred = self(x)

        # compute loss
        loss = self.loss_fn(y_pred, y[:, 0, ...])

        # compute iou
        mIoU = 0
        y_pred = torch.softmax(y_pred, dim=1)
        for c in range(y_pred.size(1)):
            mIoU += iou(y_pred[:, c:c+1, ...], y == c, w=(y != 255))
        mIoU /= y_pred.size(1)
        self.log('val/mIoU', mIoU, prog_bar=True)
        self.log('val/loss', loss)

        # log images
        if batch_idx == self.val_batch_id:
            self._log_predictions(x, y, y_pred, prefix='val')

        return loss

    def test_step(self, batch, batch_idx):

        # transfer to suitable device and get labels
        x, y = batch

        # forward prop
        y_pred = self(x)

        # compute loss
        loss = self.loss_fn(y_pred, y[:, 0, ...])

        # compute iou
        mIoU = 0
        y_pred = torch.softmax(y_pred, dim=1)
        for c in range(y_pred.size(1)):
            mIoU += iou(y_pred[:, c:c+1, ...], y == c, w=(y != 255))
        mIoU /= y_pred.size(1)
        self.log('test/mIoU', mIoU, prog_bar=True)
        self.log('test/loss', loss)

        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        scheduler = ReduceLROnPlateau(optimizer, 'max')
        return {"optimizer": optimizer}
        # return {"optimizer": optimizer, "lr_scheduler": scheduler, "monitor": 'val/mIoU'}

    def on_epoch_start(self):
        set_seed(rnd.randint(100000))
        self.train_batch_id = rnd.randint(self.trainer.num_training_batches)
        if len(self.trainer.num_val_batches) > 0:
            self.val_batch_id = rnd.randint(self.trainer.num_val_batches[0])

    def on_epoch_end(self):
        torch.save(self.state_dict(), os.path.join(self.logger.log_dir, 'checkpoints', 'epoch=%d-step=%d.ckpt' %
                                                   (self.current_epoch, self.global_step)))

    def segment(self, data, input_shape, in_channels=1, batch_size=1, step_size=None, train=False, track_progress=False,
                device=0, orientations=(0,), normalization='unit'):

        segmentation = segment(data, self, input_shape[1:], in_channels=in_channels, batch_size=batch_size,
                               step_size=step_size, train=train, track_progress=track_progress, device=device,
                               orientations=orientations, normalization=normalization)

        return segmentation

    def validate(self, data, labels, input_shape, in_channels=1, classes_of_interest=(0, 1), batch_size=1,
                 write_dir=None, val_file=None, track_progress=False, device=0, orientations=(0,), normalization='unit',
                 hausdorff=False, report=True):

        js, ams = validate(self, data, labels, input_shape[1:], in_channels=in_channels,
                           classes_of_interest=classes_of_interest, batch_size=batch_size, write_dir=write_dir,
                           val_file=val_file, track_progress=track_progress, device=device, orientations=orientations,
                           normalization=normalization, hausdorff=hausdorff, report=report)

        return np.mean(js)

    def _log_predictions(self, x, y, y_pred, prefix='train'):

        # get the tensorboard summary writer
        tensorboard = self.logger.experiment

        # select the center slice if 3D
        if x.ndim == 4:  # 2D data
            x = x[:, x.size(1) // 2, :, :]
            y = y[:, 0, :, :]
            y_ = np.argmax(y_pred.detach().cpu().numpy(), axis=1)
        else:  # 3D data
            s = x.size(2) // 2
            x = x[:, 0, s, :, :]
            y = y[:, 0, s, :, :]
            y_ = np.argmax(y_pred[:, :, s].detach().cpu().numpy(), axis=1)

        # overlay the image with the labels (boundary) and predictions (pixel-wise)
        x_ = np.zeros((x.size(0), x.size(1), x.size(2), 3))
        for b in range(x.size(0)):
            tmp = overlay(x[b].cpu().numpy(), y[b].cpu().numpy(), colors=COLORS, boundaries=True)
            x_[b] = overlay(tmp, y_[b], colors=COLORS)

        tensorboard.add_images(prefix + '/prediction', x_, global_step=self.current_epoch, dataformats='NHWC')


class UNet2D(UNet):

    def __init__(self, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1), feature_maps=64, levels=4,
                 skip_connections=True, residual_connections=False, norm='instance', activation='relu', dropout_enc=0.0,
                 dropout_dec=0.0, loss_fn='ce', lr=1e-3):
        super().__init__(input_shape=input_shape, in_channels=in_channels, coi=coi, feature_maps=feature_maps,
                         levels=levels, skip_connections=skip_connections, residual_connections=residual_connections,
                         norm=norm, activation=activation, dropout_enc=dropout_enc, dropout_dec=dropout_dec,
                         loss_fn=loss_fn, lr=lr)

        # contractive path
        self.encoder = UNetEncoder2D(self.in_channels, feature_maps=self.feature_maps, levels=self.levels,
                                     norm=self.norm, dropout=self.dropout_enc, activation=self.activation)
        # expansive path
        self.decoder = UNetDecoder2D(self.out_channels, feature_maps=self.feature_maps, levels=self.levels,
                                     skip_connections=self.skip_connections,
                                     residual_connections=self.residual_connections, norm=self.norm,
                                     dropout=self.dropout_dec, activation=self.activation)


class UNet3D(UNet):

    def __init__(self, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1), feature_maps=64, levels=4,
                 skip_connections=True, residual_connections=False, norm='instance', activation='relu', dropout_enc=0.0,
                 dropout_dec=0.0, loss_fn='ce', lr=1e-3):
        super().__init__(input_shape=input_shape, in_channels=in_channels, coi=coi, feature_maps=feature_maps,
                         levels=levels, skip_connections=skip_connections, residual_connections=residual_connections,
                         norm=norm, activation=activation, dropout_enc=dropout_enc, dropout_dec=dropout_dec,
                         loss_fn=loss_fn, lr=lr)

        # contractive path
        self.encoder = UNetEncoder3D(self.in_channels, feature_maps=self.feature_maps, levels=self.levels,
                                     norm=self.norm, dropout=self.dropout_enc, activation=self.activation)
        # expansive path
        self.decoder = UNetDecoder3D(self.out_channels, feature_maps=self.feature_maps, levels=self.levels,
                                     skip_connections=self.skip_connections,
                                     residual_connections=self.residual_connections, norm=self.norm,
                                     dropout=self.dropout_dec, activation=self.activation)


class DenseUNet2D(UNet):

    def __init__(self, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1), feature_maps=64, levels=4,
                 skip_connections=True, residual_connections=False, norm='instance', activation='relu', dropout_enc=0.0,
                 dropout_dec=0.0, num_layers=4, k=16, bn_size=2, loss_fn='ce', lr=1e-3):
        super().__init__(input_shape=input_shape, in_channels=in_channels, coi=coi, feature_maps=feature_maps,
                         levels=levels, skip_connections=skip_connections, residual_connections=residual_connections,
                         norm=norm, activation=activation, dropout_enc=dropout_enc, dropout_dec=dropout_dec,
                         loss_fn=loss_fn, lr=lr)

        # parameters
        self.num_layers = int(num_layers)
        self.k = int(k)
        self.bn_size = float(bn_size)

        # contractive path
        self.encoder = DenseUNetEncoder2D(self.in_channels, feature_maps=self.feature_maps, levels=self.levels,
                                          norm=self.norm, dropout=self.dropout_enc, activation=self.activation,
                                          num_layers=self.num_layers, k=self.k, bn_size=self.bn_size)
        # expansive path
        self.decoder = DenseUNetDecoder2D(self.out_channels, feature_maps=self.feature_maps, levels=self.levels,
                                          skip_connections=self.skip_connections,
                                          residual_connections=self.residual_connections, norm=self.norm,
                                          dropout=self.dropout_dec, activation=self.activation,
                                          num_layers=self.num_layers, k=self.k, bn_size=self.bn_size)


class DenseUNet3D(UNet):

    def __init__(self, input_shape=(1, 256, 256), in_channels=1, coi=(0, 1), feature_maps=64, levels=4,
                 skip_connections=True, residual_connections=False, norm='instance', activation='relu', dropout_enc=0.0,
                 dropout_dec=0.0, num_layers=4, k=16, bn_size=2, loss_fn='ce', lr=1e-3):
        super().__init__(input_shape=input_shape, in_channels=in_channels, coi=coi, feature_maps=feature_maps,
                         levels=levels, skip_connections=skip_connections, residual_connections=residual_connections,
                         norm=norm, activation=activation, dropout_enc=dropout_enc, dropout_dec=dropout_dec,
                         loss_fn=loss_fn, lr=lr)

        # parameters
        self.num_layers = int(num_layers)
        self.k = int(k)
        self.bn_size = float(bn_size)

        # contractive path
        self.encoder = DenseUNetEncoder3D(self.in_channels, feature_maps=self.feature_maps, levels=self.levels,
                                          norm=self.norm, dropout=self.dropout_enc, activation=self.activation,
                                          num_layers=self.num_layers, k=self.k, bn_size=self.bn_size)
        # expansive path
        self.decoder = DenseUNetDecoder3D(self.out_channels, feature_maps=self.feature_maps, levels=self.levels,
                                          skip_connections=self.skip_connections,
                                          residual_connections=self.residual_connections, norm=self.norm,
                                          dropout=self.dropout_dec, activation=self.activation,
                                          num_layers=self.num_layers, k=self.k, bn_size=self.bn_size)



import numpy as np
from skimage.measure import label
from skimage.segmentation import mark_boundaries


COLORS = [
    (1, 0, 0),
    (0, 1, 0),
    (0, 0, 1),
    (1, 1, 0),
    (0, 1, 1),
    (1, 0, 1)
]


def overlay(x, y, colors=None, alpha=0.2, boundaries=False):
    """
    Overlay an image with a segmentation map

    :param x: input (grayscale) image
    :param y: label map (all zeros are assumed background)
    :param colors: a list of 3-tuples specifying the colors that correspond to the different labels (randomly chosen if not specified)
    :param alpha: opacity of the label overlay
    :param boundaries: boolean that specifies whether a boundary map should be extracted prior to the overlay
    :return: a color image with the labels overlayed
    """
    y = y.astype('uint8')

    # if colors are not specific, generate random colors
    if colors is None:
        colors = []
        for l in range(np.max(y)):
            colors.append(tuple(np.random.rand(3)))

    # boundary case
    if boundaries:
        if x.ndim == 3 and x.shape[2] == 3:
            seg_img = x
        else:
            seg_img = np.stack((x, x, x), axis=2)
        for l in range(np.max(y)):
            seg_img = mark_boundaries(seg_img, label((y == (l + 1)).astype('uint8')), mode='thick', color=colors[l])

    # segment case
    else:
        if x.ndim == 3 and x.shape[2] == 3:
            seg_img = x
            r = np.zeros_like(x[..., 0])
            g = np.zeros_like(x[..., 1])
            b = np.zeros_like(x[..., 2])
            w1 = np.ones_like(x[..., 0]) * (1 - alpha)
            w2 = np.ones_like(x[..., 0]) * alpha
        else:
            seg_img = np.stack((x, x, x), axis=2)
            r = np.zeros_like(x)
            g = np.zeros_like(x)
            b = np.zeros_like(x)
            w1 = np.ones_like(x) * (1 - alpha)
            w2 = np.ones_like(x) * alpha

        w1[y == 0] = 1
        w1 = np.stack((w1, w1, w1), axis=2)
        w2[y == 0] = 0
        w2 = np.stack((w2, w2, w2), axis=2)

        for l in range(np.max(y)):
            for d in range(3):
                r[y == (l + 1)] = colors[l][0]
                g[y == (l + 1)] = colors[l][1]
                b[y == (l + 1)] = colors[l][2]

        overlay = np.stack((r, g, b), axis=2)
        seg_img = w1 * seg_img + w2 * overlay

    return seg_img



import random

import numpy as np
import torch
import torchvision.utils as vutils

from neuralnets.util.io import read_volume, mkdir

from scipy.ndimage.morphology import binary_opening
from torch.utils.tensorboard import SummaryWriter


def sample_labeled_input(data, labels, input_shape, zyx=None, preloaded=True, type='pngseq', data_shape=None):
    """
    Generate an input and target sample of certain shape from a labeled dataset

    :param data: data to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param labels: labels to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param input_shape: (z, x, y) shape of the sample
    :param zyx: (z, x, y) location to be sample (randomly sampled if not provided)
    :param preloaded: boolean that specifies whether the data is already in RAM
    :param type: type of the dataset that should be loaded in RAM (only necessary if preloaded==False)
    :param data_shape: (z, x, y) shape of the dataset to sample from (only necessary if preloaded==False)
    :return: a random sample
    """

    # extract input and target patch
    if preloaded:  # if preloaded, we can simply load it from RAM
        # generate random position
        if zyx is None:
            z = np.random.randint(0, data.shape[0] - input_shape[0] + 1)
            y = np.random.randint(0, data.shape[1] - input_shape[1] + 1)
            x = np.random.randint(0, data.shape[2] - input_shape[2] + 1)
        else:
            z, y, x = zyx

        input = data[z:z + input_shape[0], y:y + input_shape[1], x:x + input_shape[2]]
        target = labels[z:z + input_shape[0], y:y + input_shape[1], x:x + input_shape[2]]
    else:  # if not preloaded, we have to additionally load it in RAM
        # generate random position
        if zyx is None:
            z = np.random.randint(0, data_shape[0] - input_shape[0] + 1)
            y = np.random.randint(0, data_shape[1] - input_shape[1] + 1)
            x = np.random.randint(0, data_shape[2] - input_shape[2] + 1)
        else:
            z, y, x = zyx

        input = read_volume(data, type=type, start=z, stop=z + input_shape[0])
        target = read_volume(labels, type=type, start=z, stop=z + input_shape[0])
        input = input[:, y:y + input_shape[1], x:x + input_shape[2]]
        target = target[:, y:y + input_shape[1], x:x + input_shape[2]]

    return input.copy(), target.copy()


def sample_unlabeled_input(data, input_shape, zyx=None, preloaded=True, type='pngseq', data_shape=None):
    """
    Generate an input sample of certain shape from an unlabeled dataset

    :param data: data to sample from (a 3D numpy array if preloaded, a directory containing the data else)
    :param input_shape: (z, x, y) shape of the sample
    :param zyx: (z, x, y) location to be sample (randomly sampled if not provided)
    :param preloaded: boolean that specifies whether the data is already in RAM
    :param type: type of the dataset that should be loaded in RAM (only necessary if preloaded==False)
    :param data_shape: (z, x, y) shape of the dataset to sample from (only necessary if preloaded==False)
    :return: a random sample
    """

    # extract input and target patch
    if preloaded:  # if preloaded, we can simply load it from RAM
        # generate random position
        if zyx is None:
            z = np.random.randint(0, data.shape[0] - input_shape[0] + 1)
            y = np.random.randint(0, data.shape[1] - input_shape[1] + 1)
            x = np.random.randint(0, data.shape[2] - input_shape[2] + 1)
        else:
            z, y, x = zyx

        input = data[z:z + input_shape[0], y:y + input_shape[1], x:x + input_shape[2]]
    else:  # if not preloaded, we have to additionally load it in RAM
        # generate random position
        if zyx is None:
            z = np.random.randint(0, data_shape[0] - input_shape[0] + 1)
            y = np.random.randint(0, data_shape[1] - input_shape[1] + 1)
            x = np.random.randint(0, data_shape[2] - input_shape[2] + 1)
        else:
            z, y, x = zyx

        input = read_volume(data, type=type, start=z, stop=z + input_shape[0])
        input = input[:, y:y + input_shape[1], x:x + input_shape[2]]

    return input.copy()


def gaussian_window(size, sigma=1):
    """
    Returns a 3D Gaussian window that can be used for window weighting and merging

    :param size: size of the window
    :param sigma: standard deviation of the gaussian
    :return: the Gaussian window
    """
    # half window sizes
    hwz = size[0] // 2
    hwy = size[1] // 2
    hwx = size[2] // 2

    # construct mesh grid
    if size[0] % 2 == 0:
        axz = np.arange(-hwz, hwz)
    else:
        axz = np.arange(-hwz, hwz + 1)
    if size[1] % 2 == 0:
        axy = np.arange(-hwy, hwy)
    else:
        axy = np.arange(-hwy, hwy + 1)
    if size[2] % 2 == 0:
        axx = np.arange(-hwx, hwx)
    else:
        axx = np.arange(-hwx, hwx + 1)
    xx, zz, yy = np.meshgrid(axx, axz, axy)

    # normal distribution
    gw = np.exp(-(xx ** 2 + yy ** 2 + zz ** 2) / (2. * sigma ** 2))

    # normalize so that the mask integrates to 1
    gw = gw / np.sum(gw)

    return gw


def parse_params(params):
    """
    Parse a YAML parameter dictionary

    :param params: dictionary containing the parameters
    :return: parsed dictionary
    """

    keys = params.keys()

    if 'input_size' in keys:
        params['input_size'] = [int(item) for item in params['input_size'].split(',')]
    if 'coi' in keys:
        params['coi'] = [int(c) for c in params['coi'].split(',')]
    if 'orientations' in keys:
        if type(params['orientations']) is int:
            params['orientations'] = [params['orientations']]
        else:
            params['orientations'] = [int(c) for c in params['orientations'].split(',')]
    if 'gpus' in keys:
        if type(params['gpus']) is int:
            params['gpus'] = [params['gpus']]
        else:
            params['gpus'] = [int(c) for c in params['gpus'].split(',')]
    if 'log_refresh_rate' in keys:
        params['log_refresh_rate'] = None if params['log_refresh_rate'] < 0 else int(params['log_refresh_rate'])
    if 'train_val_test_split' in keys:
        params['train_val_test_split'] = [float(item) for item in params['train_val_test_split'].split(',')]

    return params


def load_unet(state_dict, device=0):
    """
    Load a pretrained pytorch U-Net state dict

    :param state_dict: state dict of a U-Net
    :param device: index of the device (if there are no GPU devices, it will be moved to the CPU)
    :return: a module that corresponds to the trained network
    """
    from neuralnets.networks.unet import UNet2D, UNet3D

    # extract the hyperparameters of the network
    in_channels = state_dict['encoder.features.convblock1.conv1.unit.0.weight'].size(1)
    feature_maps = state_dict['encoder.features.convblock1.conv1.unit.0.weight'].size(0)
    out_channels = state_dict['decoder.output.weight'].size(0)
    coi = tuple(np.arange(0, out_channels))
    levels = int(list(state_dict.keys())[-3][len('encoder.features.convblock')])
    skip_connections = True
    norm = 'batch' if 'norm' in list(state_dict.keys())[2] else 'instance'
    activation = 'relu'
    dropout_enc = 0.0
    dropout_dec = 0.0

    # initialize the network
    if state_dict['encoder.features.convblock1.conv1.unit.0.weight'].dim() == 4:
        net = UNet2D(in_channels=in_channels, coi=coi, feature_maps=feature_maps, levels=levels,
                     skip_connections=skip_connections, norm=norm, activation=activation, dropout_enc=dropout_enc,
                     dropout_dec=dropout_dec)
    else:
        net = UNet3D(in_channels=in_channels, coi=coi, feature_maps=feature_maps, levels=levels,
                     skip_connections=skip_connections, norm=norm, activation=activation, dropout_enc=dropout_enc,
                     dropout_dec=dropout_dec)

    # load the parameters in the model
    net.load_state_dict(state_dict)

    # map to the correct device
    module_to_device(net, device=device)

    return net


def load_bvae(state_dict, device=0):
    """
    Load a pretrained pytorch BVAE state dict

    :param state_dict: state dict of a bvae
    :param device: index of the device (if there are no GPU devices, it will be moved to the CPU)
    :return: a module that corresponds to the trained network
    """
    from neuralnets.networks.bvae import BVAE

    # extract the hyperparameters of the network
    feature_maps = state_dict['encoder.features.convblock1.conv1.unit.0.weight'].size(0)
    levels = int(list(state_dict.keys())[-5][len('decoder.features.upconv')])
    bottleneck_in_features = state_dict['encoder.bottleneck.0.weight'].size(1)
    bottleneck_dim = state_dict['encoder.bottleneck.0.weight'].size(0) // 2
    x = int(np.sqrt(bottleneck_in_features * 2**(3*levels - 1) / feature_maps))
    norm = 'batch' if 'norm' in list(state_dict.keys())[2] else 'instance'
    beta = 0.0
    activation = 'relu'
    dropout_enc = 0.0
    dropout_dec = 0.0

    # initialize the network
    net = BVAE(beta=beta, input_size=[x, x], bottleneck_dim=bottleneck_dim, feature_maps=feature_maps,
               levels=levels, dropout_enc=dropout_enc, dropout_dec=dropout_dec, norm=norm, activation=activation)

    # load the parameters in the model
    net.load_state_dict(state_dict)

    # map to the correct device
    module_to_device(net, device=device)

    return net


def load_net(model_file, device=0):
    """
    Load a pretrained pytorch network, currently only support for U-Net and BVAE

    :param model_file: path to the state dict checkpoint
    :param device: index of the device (if there are no GPU devices, it will be moved to the CPU)
    :return: a module that corresponds to the trained network
    """

    # load the state dict to the correct device
    map_location = 'cpu' if not torch.cuda.is_available() or device == 'cpu' else 'cuda:' + str(device)
    state_dict = torch.load(model_file, map_location=map_location)

    # get parameters of the state dict to identify the network
    out_channels = state_dict['decoder.output.weight'].size(0)

    net = None
    if out_channels == 1:  # autoencoder architecture
        net = load_bvae(state_dict, device=device)
    elif out_channels > 1:  # U-Net architecture
        net = load_unet(state_dict, device=device)

    return net


def save_net(model, model_file):
    """
    Save a pytorch network

    :param model: path to the state dict checkpoint
    :param model_file: path to the state dict checkpoint
    """
    torch.save(model.state_dict(), model_file)


def set_seed(seed):
    """
    Sets the seed of all randomized modules (useful for reproducibility)

    :param seed: seed number
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def module_to_device(module, device):
    """
    Transfers a pytorch module to a specific GPU device

    :param module: module that should be transferred
    :param device: index of the device, or 'cpu' (if there are no GPU devices, it will be moved to the CPU)
    """
    if torch.cuda.is_available() and (device.__class__ == int or device.__class__ == float):
        return module.to(device=torch.device('cuda:' + str(int(device))))
    else:
        return module.cpu()


def tensor_to_device(x, device):
    """
    Transfers a pytorch tensor to a specific GPU device

    :param x: tensor or sequence/list of tensors that should be transferred
    :param device: index of the device (if there are no GPU devices or device is None, it will be moved to the CPU)
    :return x: same tensor, but switched to device
    """
    if isinstance(x, tuple) or isinstance(x, list):
        if torch.cuda.is_available() and (device.__class__ == int or device.__class__ == float):
            return [xx.to(device=torch.device('cuda:' + str(int(device)))) for xx in x]
        else:
            return [xx.cpu() for xx in x]
    else:
        if torch.cuda.is_available() and (device.__class__ == int or device.__class__ == float):
            return x.to(device=torch.device('cuda:' + str(int(device))))
        else:
            return x.cpu()


def augment_samples(data, augmenter=None):
    """
    Augment a tensor with a specific augmenter

    :param data: tensor or sequence/list of tensors that should be augmented
    :param optional augmenter: augmenter that should be used (original data is returned if this is not specified)
    :return data: augmented tensor (or list of tensors)
    """
    if augmenter is not None:
        if isinstance(data, tuple) or isinstance(data, list):
            bs = [x.size(0) for x in data]
            data = [x.float() for x in data]
            data_aug = augmenter(torch.cat(data, dim=0)).float()
            return torch.split(data_aug, bs, dim=0)
        else:
            return augmenter(data.float()).float()
    return data


def get_labels(y, coi, dtype=int):
    """
    Maps general annotated image tensors to indexed labels for particular classes of interest

    :param y: annotated image tensor (B, N_1, N_2, ...)
    :param coi: classes of interest
    :param optional dtype: type of the tensor (typically integers)
    :return: indexed label tensor, ready for use in most loss functions (B, N_1, N_2, ...)
    """
    labels = torch.zeros_like(y, dtype=dtype)

    # convert labels to integers
    y = torch.round(y.float()).long()

    # loop over classes of interest
    for i, c in enumerate(coi):
        if i > 0:
            labels[y == c] = i

    # check if other classes are annotated, these can be labeled as background
    for c in torch.unique(y):
        if not c in coi and not c == 255:
            labels[y == c] = 0
    return labels


def get_unlabeled(y, dtype=int):
    """
    Maps general annotated image tensors to an indexed image of unlabeled pixels

    :param y: annotated image tensor (B, N_1, N_2, ...)
    :param coi: classes of interest
    :param optional dtype: type of the tensor (typically integers)
    :return: indexed label tensor, ready for use in most loss functions (B, N_1, N_2, ...)
    """
    unlabeled = torch.zeros_like(y, dtype=dtype)
    unlabeled[y == 255] = 1
    return unlabeled


def log_scalars(scalars, names, writer, epoch=0):
    """
    Writes a list of scalars to a tensorboard events file

    :param scalars: list of scalars (can be tensors or numpy arrays) that should be logged
    :param names: list of names that correspond to the scalars
    :param writer: writer used for logging
    :param epoch: current epoch
    """
    for name, scalar in zip(names, scalars):
        writer.add_scalar(name, scalar, epoch)


def log_images_2d(images, names, writer, epoch=0, scale_each=True):
    """
    Writes a list of 2D images to a tensorboard events file

    :param images: list of (2D) images (in pytorch tensor format, size [B, {1,3}, Y, X]) that should be logged
    :param names: list of names that correspond to the images
    :param writer: writer used for logging
    :param optional epoch: current epoch
    :param optional scale_each: scale each image or not
    """
    for id, x in zip(names, images):
        x = vutils.make_grid(x, normalize=x.max() - x.min() > 0, scale_each=scale_each)
        writer.add_image(id, x, epoch)


def log_images_3d(images, names, writer, epoch=0, scale_each=True):
    """
    Writes a list of 3D images to a tensorboard events file.
    For efficiency reasons, the center z-slice is selected from the 3D image

    :param images: list of (3D) images (in pytorch tensor format, size [B, {1,3}, Z, Y, X]) that should be logged
    :param names: list of names that correspond to the images
    :param writer: writer used for logging
    :param epoch: current epoch
    :param optional scale_each: scale each image or not
    """
    for id, x in zip(names, images):
        x = x[:, :, x.size(2) // 2, :, :]
        x = vutils.make_grid(x, normalize=x.max() - x.min() > 0, scale_each=scale_each)
        writer.add_image(id, x, epoch)


def log_hparams(gs, log_dir='logs'):
    mkdir(log_dir)
    params = gs.cv_results_['params']
    metrics = gs.cv_results_['mean_test_score']
    for j, p_dict in enumerate(params):
        tb = SummaryWriter(log_dir=log_dir)
        p_dict = {key: (str(p_dict[key]) if (isinstance(p_dict[key], list) or isinstance(p_dict[key], tuple))
                        else p_dict[key]) for key in list(p_dict.keys())}
        m_dict = {'mIoU': metrics[j]}
        tb.add_hparams(p_dict, m_dict)
        tb.close()


def clean_labels(y, n_classes):
    y_clean = y
    y = y.data.cpu().numpy()
    for b in range(y.shape[0]):
        y_b = y_clean[b, 0, ...]
        for c in range(n_classes):
            if not (c == 0 or c == y.shape[1] - 1):
                mask = binary_opening(y[b, 0, ...] == c)
                y_b[y_b == c] = 0
                y_b[torch.Tensor(mask).bool()] = c
        y_clean[b, 0, ...] = y_b
    return y_clean


def normalize(x, type=None, factor=None, mu=None, sigma=None):
    """
    Normalizes an numpy array

    :param x: an arbitrary numpy array
    :param type: the desired type of normalization (None, z, unit or minmax)
    :param factor: normalization factor (only if type is unit)
    :param mu: normalization mean (only if type is z)
    :param sigma: normalization std (only if type is z)
    :return: the normalized numpy array
    """
    if type == 'z':
        # apply z normalization
        if mu is None:
            mu = 0
        if sigma is None:
            sigma = 1
        return (x - mu) / sigma
    elif type == 'minmax':
        m = x.min()
        M = x.max()
        eps = 1e-5
        return (x - m + eps) / (M - m + eps)
    else:
        # apply unit normalization
        if factor == None:
            factors = {np.dtype('int8'): 2 ** 8 - 1,
                       np.dtype('uint8'): 2 ** 8 - 1,
                       np.dtype('int16'): 2 ** 16 - 1,
                       np.dtype('uint16'): 2 ** 16 - 1,
                       np.dtype('int32'): 2 ** 32 - 1,
                       np.dtype('uint32'): 2 ** 32 - 1,
                       np.dtype('int64'): 2 ** 64 - 1,
                       np.dtype('uint64'): 2 ** 64 - 1}
            if x.dtype in factors:
                factor = factors[x.dtype]
            else:
                factor = 1.
        return x / factor


def _find_split(y, test_size=0.33, axis=0):

    # reorient data
    if axis == 1:
        y_ = np.transpose(y, axes=(1, 0, 2))
    elif axis == 2:
        y_ = np.transpose(y, axes=(2, 1, 0))
    else:
        y_ = y

    start = 0
    stop = y_.shape[0] - 1
    n_total = np.sum(y_ != 255)

    while start != stop:

        # split halfway between start and stop
        split = (start + stop) // 2

        # compute test size
        n_test = np.sum(y_[split:] != 255)
        split_test_size = n_test / n_total

        if split == start:
            break

        if split_test_size < test_size:
            # test set needs to be larger
            stop = split
        else:
            # test set needs to be smaller
            start = split

    split = start

    return split


def _maximize_isotropy(shape, splits):

    # for each split and for both train and test volumes,
    # we compute the difference between a cubic hull and the actual volume
    maximum_covered_fraction = 0
    best_dim = 0
    for d, split in enumerate(splits):

        # volume train set
        train_shape = np.asarray(shape)
        train_shape[d] = split
        train_volume = np.prod(train_shape)
        train_cubic_hull_volume = np.max(train_shape)**3

        # volume test set
        test_shape = np.asarray(shape)
        test_shape[d] = split
        test_volume = np.prod(test_shape)
        test_cubic_hull_volume = np.max(test_shape)**3

        # maximize covered fraction
        covered_fraction = train_volume / train_cubic_hull_volume + test_volume / test_cubic_hull_volume
        if covered_fraction > maximum_covered_fraction:
            maximum_covered_fraction = covered_fraction
            best_dim = d

    return best_dim, splits[best_dim]


def train_test_split(x, y=None, test_size=0.33):
    """
    Splits a (both labeled and unlabeled) data in a stratified train and test set while maximizing isotropic data dimensions

    :param x: input data
    :param y: corresponding labels, optional
    :param test_size: relative size of the test set
    :return: train test splits of the data
    """

    if y is None:  # unlabeled dataset, easier case
        splits = [int(sz * (1-test_size)) for sz in x.shape]
        d, s = _maximize_isotropy(x.shape, splits)
        x_train, x_test = np.split(x, [s], d)
        return x_train, x_test
    else:  # labeled dataset, additionally make sure labels are balanced in train/test folds
        splits = [_find_split(y, test_size=test_size, axis=axis) for axis in [0, 1, 2]]
        d, s = _maximize_isotropy(y.shape, splits)
        x_train, x_test = np.split(x, [s], d)
        y_train, y_test = np.split(y, [s], d)
        return x_train, y_train, x_test, y_test




"""
    This is a script that illustrates training a beta variational autoencoder
"""

"""
    Necessary libraries
"""
import argparse
import os

import torch.optim as optim
from torch.utils.data import DataLoader
from torchvision.transforms import Compose

from neuralnets.data.datasets import UnlabeledVolumeDataset
from neuralnets.networks.bvae import BVAE
from neuralnets.util.augmentation import *
from neuralnets.util.io import print_frm
from neuralnets.util.tools import set_seed

"""
    Parse all the arguments
"""
print_frm('Parsing arguments')
parser = argparse.ArgumentParser()

# logging parameters
parser.add_argument("--seed", help="Seed for randomization", type=int, default=0)
parser.add_argument("--device", help="GPU device for computations", type=int, default=0)
parser.add_argument("--log_dir", help="Logging directory", type=str, default="bvae")
parser.add_argument("--print_stats", help="Number of iterations between each time to log training losses",
                    type=int, default=50)

# network parameters
parser.add_argument("--data_dir", help="Data directory", type=str, default="../../../data")
parser.add_argument("--input_size", help="Size of the blocks that propagate through the network",
                    type=str, default="64,64")
parser.add_argument("--fm", help="Number of initial feature maps in the segmentation U-Net", type=int, default=128)
parser.add_argument("--levels", help="Number of levels in the segmentation U-Net (i.e. number of pooling stages)",
                    type=int, default=3)
parser.add_argument("--dropout", help="Dropout", type=float, default=0.0)
parser.add_argument("--norm", help="Normalization in the network (batch or instance)", type=str, default="batch")
parser.add_argument("--activation", help="Non-linear activations in the network", type=str, default="relu")
parser.add_argument("--beta", help="Beta value for the KL loss", type=float, default=0.0)
parser.add_argument("--bottleneck", help="Dimensionality of the bottleneck", type=int, default=128)

# optimization parameters
parser.add_argument("--lr", help="Learning rate of the optimization", type=float, default=1e-3)
parser.add_argument("--step_size", help="Number of epochs after which the learning rate should decay",
                    type=int, default=10)
parser.add_argument("--gamma", help="Learning rate decay factor", type=float, default=1)
parser.add_argument("--epochs", help="Total number of epochs to train", type=int, default=200)
parser.add_argument("--len_epoch", help="Number of iteration in each epoch", type=int, default=100)
parser.add_argument("--test_freq", help="Number of epochs between each test stage", type=int, default=1)
parser.add_argument("--train_batch_size", help="Batch size in the training stage", type=int, default=4)
parser.add_argument("--test_batch_size", help="Batch size in the testing stage", type=int, default=4)

args = parser.parse_args()
args.input_size = [int(item) for item in args.input_size.split(',')]

"""
Fix seed (for reproducibility)
"""
set_seed(args.seed)

"""
    Setup logging directory
"""
print_frm('Setting up log directories')
if not os.path.exists(args.log_dir):
    os.mkdir(args.log_dir)

"""
    Load the data
"""
input_shape = (1, args.input_size[0], args.input_size[1])
print_frm('Loading data')
augmenter = Compose(
    [ToFloatTensor(device=args.device), Rotate90(), FlipX(prob=0.5), FlipY(prob=0.5), ContrastAdjust(adj=0.1),
     RandomDeformation_2D(input_shape[1:], grid_size=(64, 64), sigma=0.01, device=args.device)])
train = UnlabeledVolumeDataset(os.path.join(args.data_dir, 'EM/EPFL/train'), input_shape=input_shape,
                               len_epoch=args.len_epoch, type='pngseq')
test = UnlabeledVolumeDataset(os.path.join(args.data_dir, 'EM/EPFL/test'), input_shape=input_shape,
                              len_epoch=args.len_epoch, type='pngseq')
train_loader = DataLoader(train, batch_size=args.train_batch_size)
test_loader = DataLoader(test, batch_size=args.train_batch_size)

"""
    Build the network
"""
print_frm('Building the network')
net = BVAE(beta=args.beta, input_size=args.input_size, bottleneck_dim=args.bottleneck, feature_maps=args.fm,
           levels=args.levels, dropout_enc=args.dropout, dropout_dec=args.dropout, norm=args.norm,
           activation=args.activation)

"""
    Setup optimization for training
"""
print_frm('Setting up optimization for training')
optimizer = optim.Adam(net.parameters(), lr=args.lr)
scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=args.step_size, gamma=args.gamma)

"""
    Train the network
"""
print_frm('Starting training')
net.train_net(train_loader, test_loader, optimizer, args.epochs, scheduler=scheduler, augmenter=augmenter,
              print_stats=args.print_stats, log_dir=args.log_dir, device=args.device)

print_frm('Finished!')



import os

import numpy as np
import torch
import torch.nn.functional as F
from progress.bar import Bar

from neuralnets.util.io import read_volume, write_volume, print_frm, mkdir
from neuralnets.util.numpy.metrics import iou, accuracy_metrics, hausdorff_distance, _iou_part_metrics, \
    _conf_matrix_metrics
from neuralnets.util.tools import gaussian_window, tensor_to_device, module_to_device, normalize


def sliding_window_multichannel(image, step_size, window_size, in_channels=1, track_progress=False,
                                normalization='unit'):
    """
    Iterator that acts as a sliding window over a multichannel 3D image

    :param image: multichannel image (4D array)
    :param step_size: step size of the sliding window (3-tuple)
    :param window_size: size of the window (3-tuple)
    :param in_channels: amount of subsequent slices that serve as input for the network (should be odd)
    :param track_progress: optionally, for tracking progress with progress bar
    :param normalization: type of data normalization (unit, z or minmax)
    """

    # adjust z-channels if necessary
    window_size = np.asarray(window_size)
    is2d = window_size[0] == 1
    if is2d:  # 2D
        window_size[0] = in_channels

    # define range
    zrange = [0]
    while zrange[-1] < image.shape[1] - window_size[0]:
        zrange.append(zrange[-1] + step_size[0])
    zrange[-1] = image.shape[1] - window_size[0]
    yrange = [0]
    while yrange[-1] < image.shape[2] - window_size[1]:
        yrange.append(yrange[-1] + step_size[1])
    yrange[-1] = image.shape[2] - window_size[1]
    xrange = [0]
    while xrange[-1] < image.shape[3] - window_size[2]:
        xrange.append(xrange[-1] + step_size[2])
    xrange[-1] = image.shape[3] - window_size[2]

    # loop over the range
    if track_progress:
        bar = Bar('Progress', max=len(zrange) * len(yrange) * len(xrange))
    for z in zrange:
        for y in yrange:
            for x in xrange:

                # yield the current window
                if is2d:
                    input = image[0, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]]
                else:
                    input = image[:, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]]
                    yield (z, y, x, image[:, z:z + window_size[0], y:y + window_size[1], x:x + window_size[2]])
                input = normalize(input, type=normalization)
                yield (z, y, x, input)

                if track_progress:
                    bar.next()
    if track_progress:
        bar.finish()


def _init_step_size(step_size, input_shape, is2d):
    if step_size == None:
        if is2d:
            step_size = (1, max(1, input_shape[0] // 2), max(1, input_shape[1] // 2))
        else:
            step_size = (max(1, input_shape[0] // 2), max(1, input_shape[1] // 2), max(1, input_shape[2] // 2))
    return step_size


def _init_gaussian_window(input_shape, is2d):
    if is2d:
        g_window = gaussian_window((1, input_shape[0], input_shape[1]), sigma=input_shape[-1] / 4)
    else:
        g_window = gaussian_window(input_shape, sigma=input_shape[-1] / 4)
    return g_window


def _init_sliding_window(data, step_size, input_shape, in_channels, is2d, track_progress, normalization):
    if is2d:
        sw = sliding_window_multichannel(data, step_size=step_size, window_size=(1, input_shape[0], input_shape[1]),
                                         in_channels=in_channels, track_progress=track_progress,
                                         normalization=normalization)
    else:
        sw = sliding_window_multichannel(data, step_size=step_size, window_size=input_shape,
                                         track_progress=track_progress, normalization=normalization)
    return sw


def _orient(data, orientation=0):
    """
    This function essentially places the desired orientation axis to that of the original Z-axis
    For example:
          (C, Z, Y, X) -> (C, Y, Z, X) for orientation=1
          (C, Z, Y, X) -> (C, X, Y, Z) for orientation=2
    Note that applying this function twice corresponds to the identity transform

    :param data: assumed to be of shape (C, Z, Y, X)
    :param orientation: 0, 1 or 2 (respectively for Z, Y or X axis)
    :return: reoriented dataset
    """
    if orientation == 1:
        return np.transpose(data, axes=(0, 2, 1, 3))
    elif orientation == 2:
        return np.transpose(data, axes=(0, 3, 2, 1))
    else:
        return data


def _pad(data, input_shape, in_channels):
    # pad data if input shape is larger than data
    in_shape = input_shape if len(input_shape) == 3 else (1, input_shape[0], input_shape[1])
    pad_width = [[0, 0], None, None, None]
    for d in range(3):
        padding = np.maximum(0, in_shape[d] - data.shape[d + 1])
        before = padding // 2
        after = padding - before
        pad_width[d + 1] = [before, after]

    # pad z-slices if necessary (required if the network uses more than 1 input channel)
    if in_channels > 1:
        c = (in_channels // 2)
        pad_width[1][0] = pad_width[1][0] + c
        pad_width[1][1] = pad_width[1][1] + c

    return np.pad(data, pad_width=pad_width, mode='symmetric'), pad_width


def _crop(data, seg_cum, counts_cum, pad_width):
    return data[:, pad_width[1][0]:data.shape[1] - pad_width[1][1], pad_width[2][0]:data.shape[2] - pad_width[2][1],
           pad_width[3][0]:data.shape[3] - pad_width[3][1]], \
           seg_cum[:, pad_width[1][0]:data.shape[1] - pad_width[1][1], pad_width[2][0]:data.shape[2] - pad_width[2][1],
           pad_width[3][0]:data.shape[3] - pad_width[3][1]], \
           counts_cum[pad_width[1][0]:data.shape[1] - pad_width[1][1], pad_width[2][0]:data.shape[2] - pad_width[2][1],
           pad_width[3][0]:data.shape[3] - pad_width[3][1]]


def _forward_prop(net, x):
    outputs = net(x)
    # if the outputs are a tuple, take the last
    if type(outputs) is tuple:
        outputs = outputs[-1]
    return F.softmax(outputs, dim=1).detach().cpu().numpy()


def _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, in_channels,
                           is2d):
    c = in_channels // 2
    for b in range(batch_size):
        (z_b, y_b, x_b) = positions[b, :]
        # take into account the gaussian filtering
        if is2d:
            z_b += c  # correct channel shift
            seg_cum[:, z_b, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += \
                np.multiply(g_window, outputs[b, ...])
            counts_cum[z_b:z_b + 1, y_b:y_b + input_shape[0], x_b:x_b + input_shape[1]] += g_window
        else:
            seg_cum[:, z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += \
                np.multiply(g_window, outputs[b, ...])
            counts_cum[z_b:z_b + input_shape[0], y_b:y_b + input_shape[1], x_b:x_b + input_shape[2]] += g_window


def _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, in_channels,
                   is2d):
    # convert to tensors and switch to correct device
    inputs = tensor_to_device(torch.FloatTensor(batch), device)
    # forward prop
    outputs = _forward_prop(net, inputs)
    # cumulate segmentation volume
    _cumulate_segmentation(seg_cum, counts_cum, outputs, g_window, positions, batch_size, input_shape, in_channels,
                           is2d)


def segment_multichannel_2d(data, net, input_shape, batch_size=1, step_size=None, train=False, track_progress=False,
                            device=0, normalization='unit'):
    """
    Segment a multichannel 2D image using a specific network

    :param data: 3D array (C, Y, X) representing the multichannel 2D image
    :param net: image-to-image segmentation network
    :param input_shape: size of the inputs (2-tuple)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param normalization: type of data normalization (unit, z or minmax)
    :return: the segmented image
    """

    # make sure we compute everything on the correct device
    module_to_device(net, device)

    # set the network in the correct mode
    if train:
        net.train()
    else:
        net.eval()

    # pad data if necessary
    data, pad_width = _pad(data[:, np.newaxis, ...], input_shape, 1)
    data = data[:, 0, ...]

    # get the amount of channels
    channels = data.shape[0]

    # initialize the step size
    step_size = _init_step_size(step_size, input_shape, True)

    # gaussian window for smooth block merging
    g_window = _init_gaussian_window(input_shape, True)

    # allocate space
    seg_cum = np.zeros((net.out_channels, 1, *data.shape[1:]))
    counts_cum = np.zeros((1, *data.shape[1:]))

    # define sliding window
    sw = _init_sliding_window(data[np.newaxis, ...], [channels, *step_size[1:]], input_shape, channels, True,
                              track_progress, normalization)

    # start prediction
    batch_counter = 0
    batch = np.zeros((batch_size, channels, *input_shape))
    positions = np.zeros((batch_size, 3), dtype=int)
    for (z, y, x, inputs) in sw:

        # fill batch
        batch[batch_counter, ...] = inputs
        positions[batch_counter, :] = [z, y, x]

        # increment batch counter
        batch_counter += 1

        # perform segmentation when a full batch is filled
        if batch_counter == batch_size:
            # process a single batch
            _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, 1,
                           True)

            # reset batch counter
            batch_counter = 0

    # don't forget to process the last batch
    _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, 1, True)

    # crop out the symmetric extension and compute segmentation
    data, seg_cum, counts_cum = _crop(data[:, np.newaxis, ...], seg_cum, counts_cum, pad_width)
    for c in range(net.out_channels):
        seg_cum[c, ...] = np.divide(seg_cum[c, ...], counts_cum)
    seg_cum = seg_cum[:, 0, ...]

    return seg_cum


def segment_multichannel_3d(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False,
                            track_progress=False, device=0, orientation=0, normalization='unit'):
    """
    Segment a multichannel 3D image using a specific network

    :param data: 4D array (C, Z, Y, X) representing the multichannel 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of the inputs (either 2 or 3-tuple)
    :param in_channels: amount of subsequent slices that serve as input for the network (should be odd)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientation: orientation to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :return: the segmented image
    """

    # make sure we compute everything on the correct device
    module_to_device(net, device)

    # set the network in the correct mode
    if train:
        net.train()
    else:
        net.eval()

    # orient data if necessary
    data = _orient(data, orientation)

    # pad data if necessary
    data, pad_width = _pad(data, input_shape, in_channels)

    # 2D or 3D
    is2d = len(input_shape) == 2

    # get the amount of channels
    channels = data.shape[0]
    if is2d:
        channels = in_channels

    # initialize the step size
    step_size = _init_step_size(step_size, input_shape, is2d)

    # gaussian window for smooth block merging
    g_window = _init_gaussian_window(input_shape, is2d)

    # allocate space
    seg_cum = np.zeros((net.out_channels, *data.shape[1:]))
    counts_cum = np.zeros(data.shape[1:])

    # define sliding window
    sw = _init_sliding_window(data, step_size, input_shape, in_channels, is2d, track_progress, normalization)

    # start prediction
    batch_counter = 0
    batch = np.zeros((batch_size, channels, *input_shape))
    positions = np.zeros((batch_size, 3), dtype=int)
    for (z, y, x, inputs) in sw:

        # fill batch
        batch[batch_counter, ...] = inputs
        positions[batch_counter, :] = [z, y, x]

        # increment batch counter
        batch_counter += 1

        # perform segmentation when a full batch is filled
        if batch_counter == batch_size:
            # process a single batch
            _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape,
                           in_channels, is2d)

            # reset batch counter
            batch_counter = 0

    # don't forget to process the last batch
    _process_batch(net, batch, device, seg_cum, counts_cum, g_window, positions, batch_size, input_shape, in_channels,
                   is2d)

    # crop out the symmetric extension and compute segmentation
    data, seg_cum, counts_cum = _crop(data, seg_cum, counts_cum, pad_width)
    for c in range(net.out_channels):
        seg_cum[c, ...] = np.divide(seg_cum[c, ...], counts_cum)

    # reorient data to its original orientation
    data = _orient(data, orientation)
    seg_cum = _orient(seg_cum, orientation)

    return seg_cum


def segment_multichannel(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False,
                         track_progress=False, device=0, orientation=0, normalization='unit'):
    """
    Segment a multichannel 2D or 3D image using a specific network

    :param data: 4D array (C, [Z, ]Y, X) representing the multichannel image
    :param net: image-to-image segmentation network
    :param input_shape: size of the inputs (either 2 or 3-tuple)
    :param in_channels: amount of subsequent slices that serve as input for the network (should be odd)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientation: orientation to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :return: the segmented image
    """

    if data.ndim == 4:
        return segment_multichannel_3d(data, net, input_shape, in_channels=in_channels, batch_size=batch_size,
                                       step_size=step_size, train=train, track_progress=track_progress, device=device,
                                       orientation=orientation, normalization=normalization)
    else:
        return segment_multichannel_2d(data, net, input_shape, batch_size=batch_size, step_size=step_size, train=train,
                                       track_progress=track_progress, device=device, normalization=normalization)


def _segment_z_block(z_block, net, sub_block_size, overlap_size, input_shape, in_channels=1, batch_size=1,
                     step_size=None, train=False, track_progress=False, device=0, orientations=(0,),
                     normalization='unit', bar=None):
    z, y, x = z_block.shape
    yb, xb = sub_block_size[1:]
    yo, xo = overlap_size[1:]

    seg_cum = np.zeros((net.out_channels, *z_block.shape))
    wgt_cum = np.zeros(z_block.shape)
    sigma_wgt_window = np.min(sub_block_size[1:]) / 4
    weight_wnd_ = gaussian_window((sub_block_size[0], sub_block_size[2], sub_block_size[1]), sigma=sigma_wgt_window)
    for j in range(0, y, yb):
        yb_start = np.maximum(0, j - yo)
        yb_stop = np.minimum(j + yb + yo, y)
        for k in range(0, x, xb):
            xb_start = np.maximum(0, k - xo)
            xb_stop = np.minimum(k + xb + xo, x)

            block = z_block[:, yb_start:yb_stop, xb_start:xb_stop]
            segmented_block = segment(block, net, input_shape, in_channels=in_channels, batch_size=batch_size,
                                      step_size=step_size, train=train, device=device, orientations=orientations, normalization=normalization)
            if block.shape == sub_block_size:
                weight_wnd = weight_wnd_
            else:
                weight_wnd = gaussian_window((block.shape[0], block.shape[2], block.shape[1]), sigma=sigma_wgt_window)

            for c in range(net.out_channels):
                seg_cum[c, :, yb_start:yb_stop, xb_start:xb_stop] \
                    = seg_cum[c, :, yb_start:yb_stop, xb_start:xb_stop] + weight_wnd * segmented_block[c]
            wgt_cum[:, yb_start:yb_stop, xb_start:xb_stop] = wgt_cum[:, yb_start:yb_stop, xb_start:xb_stop] + weight_wnd

            if track_progress:
                bar.next()

    for c in range(net.out_channels):
        seg_cum[c] = seg_cum[c] / wgt_cum

    return seg_cum


def _cumulate_validation_metrics(z_block_pred, z_block_labels, js_cum, ams_cum, classes_of_interest):
    all_labeled = np.sum(z_block_labels == 255) == 0
    w = None if all_labeled else z_block_labels != 255

    js = np.asarray(
        [_iou_part_metrics(z_block_labels == c, z_block_pred[i], w=w) for i, c in enumerate(classes_of_interest)])
    ams = np.asarray(
        [_conf_matrix_metrics(z_block_labels == c, z_block_pred[i], w=w) for i, c in enumerate(classes_of_interest)])

    js_cum += js
    ams_cum += ams

    return js_cum, ams_cum


def _compute_validation_metrics(js_cum, ams_cum, classes_of_interest):
    js = js_cum[:, 0] / (js_cum[:, 1] + js_cum[:, 2] - js_cum[:, 0])

    total = np.sum(ams_cum, axis=1)
    accuracy = (ams_cum[:, 0] + ams_cum[:, 1] + 1) / (total + 1)
    recall = (ams_cum[:, 0] + 1) / (ams_cum[:, 0] + ams_cum[:, 3] + 1)
    specificity = (ams_cum[:, 1] + 1) / (ams_cum[:, 1] + ams_cum[:, 3] + 1)
    balanced_accuracy = (recall + specificity) / 2
    precision = (ams_cum[:, 0] + 1) / (ams_cum[:, 0] + ams_cum[:, 3] + 1)
    f1 = 2 * (precision * recall) / (precision + recall)
    ams = np.asarray([(accuracy[i], balanced_accuracy[i], precision[i], recall[i], f1[i]) for i, c in
                      enumerate(classes_of_interest)])

    return js, ams


def _write_segmentation(segmentation, write_dir, classes_of_interest=(0, 1), index_inc=0):
    mkdir(write_dir)
    segmentation_volume = np.zeros(segmentation.shape[1:])
    for i, c in enumerate(classes_of_interest):
        segmentation_volume[segmentation[i] > 0.5] = c
    write_volume(segmentation_volume, write_dir, index_inc=index_inc, type='pngseq')


def _validate_ram(segmentation, labels, classes_of_interest=(0, 1), hausdorff=False, report=True):
    """
    Validate a segmentation based on available labels

    :param segmentation: 4D array (C, Z, Y, X) representing the a class probability distribution over the 3D image
    :param labels: 3D array (Z, Y, X) representing the 3D labels
    :param classes_of_interest: index of the label of interest
    :param hausdorff: compute hausdorff or not
    :param report: print the validation report or not (only required if labels is not None)
    :return: validation results, i.e. accuracy, precision, recall, f-score, jaccard and dice score
    """

    # compute metrics
    all_labeled = np.sum(labels == 255) == 0
    w = None if all_labeled else labels != 255
    comp_hausdorff = all_labeled and hausdorff
    js = np.asarray([iou(labels == c, segmentation[i], w=w) for i, c in enumerate(classes_of_interest)])
    ams = np.asarray([accuracy_metrics(labels == c, segmentation[i], w=w) for i, c in enumerate(classes_of_interest)])
    hs = np.zeros_like(js)
    for i, c in enumerate(classes_of_interest):
        if comp_hausdorff:
            hs[i] = hausdorff_distance(labels == c, segmentation[i])[0]
        else:
            hs[i] = -1

    # report results if necessary
    if report:
        _report_validation(js, ams, classes_of_interest, hs=hs)

    return js, ams


def _report_validation(js, ams, classes_of_interest, hs=None):
    for i, c in enumerate(classes_of_interest):
        print_frm('Validation performance for class %d: ' % c)
        print_frm('    - Accuracy: %f' % ams[i, 0])
        print_frm('    - Balanced accuracy: %f' % ams[i, 1])
        print_frm('    - Precision: %f' % ams[i, 2])
        print_frm('    - Recall: %f' % ams[i, 3])
        print_frm('    - F1 score: %f' % ams[i, 4])
        print_frm('    - IoU: %f' % js[i])
        if hs is not None:
            print_frm('    - Hausdorff distance: %f' % hs[i])

    print_frm('Validation performance mean: ')
    print_frm('    - Accuracy: %f' % np.mean(ams[:, 0]))
    print_frm('    - Balanced accuracy: %f' % np.mean(ams[:, 1]))
    print_frm('    - Precision: %f' % np.mean(ams[:, 2]))
    print_frm('    - Recall: %f' % np.mean(ams[:, 3]))
    print_frm('    - F1 score: %f' % np.mean(ams[:, 4]))
    print_frm('    - mIoU: %f' % np.mean(js))
    if hs is not None:
        print_frm('    - Hausdorff distance: %f' % np.mean(hs))


def _save_validation(js, ams, val_file):
    np.save(val_file, np.concatenate((js[:, np.newaxis], ams), axis=1))


def segment_read(data, net, input_shape, write_dir, start=0, stop=-1, block_size=(50, 1024, 1024),
                 overlap_size=(16, 128, 128), in_channels=1, batch_size=1, step_size=None, train=False,
                 track_progress=False, device=0, orientations=(0,), normalization='unit', type='pngseq', labels=None,
                 val_file=None, classes_of_interest=(0, 1), report=True):
    """
    Segment a 3D image using a specific network based on a directory of data.
    This function is designed for large datasets where RAM becomes a bottleneck. The segmentation is performed with
    overlapping blocks and optional writing/validation is also performed on a block basis. Note that the hausdorff
    distance metric is not supported du to block-based processing

    :param data: directory that contains a 3D array (Z, Y, X) representing the 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of
    :param start: start position to segment (0 for beginning of the data)
    :param stop: stop position to segment (-1 for end of the data)
    :param block_size: block size for processing
    :param overlap_size: overlap size for avoiding blocking artifacts
    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
    :param write_dir: destination for the segmentation, set this equal to None, in which case only validation will be performed
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientations: list of orientations to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :param type: type of the data (pngseq or tifseq)
    :param labels: directory that contains a 3D array (Z, Y, X) representing the 3D labels
    :param val_file: destination of the validation file (only required if labels is not None)
    :param classes_of_interest: index of the label of interest (required if labels is not None or write_dir is not None)
    :param report: print the validation report or not (only required if labels is not None)
    """

    # get the dimensions of the data
    files = os.listdir(data)
    z = len(files)
    y, x = read_volume(data, type=type, start=0, stop=1).shape[1:]

    # get the start and stop positions
    if stop < 0:
        stop = z
    z = stop - start

    # the block size
    zb, yb, xb = block_size
    zo, yo, xo = overlap_size

    # initialize the validation metrics
    js_cum = np.zeros((len(classes_of_interest), 3))
    ams_cum = np.zeros((len(classes_of_interest), 4))

    # initialize progress bar
    if track_progress:
        bar = Bar('Progress', max=len(range(0, z, zb)) * len(range(0, y, yb)) * len(range(0, x, xb)))
    else:
        bar = None

    # first check whether there are at least zb slices
    # if not, we can straightforward segment the z-block
    if z <= zb:
        z_block = read_volume(data, type=type, start=start, stop=stop)
        z_block_segmented = _segment_z_block(z_block, net, (zb, yb, xb), (zo, yo, xo), input_shape,
                                             in_channels=in_channels, batch_size=batch_size, step_size=step_size,
                                             train=train, track_progress=track_progress, device=device,
                                             orientations=orientations, normalization=normalization, bar=bar)
        if write_dir is not None:
            _write_segmentation(z_block_segmented, write_dir, classes_of_interest=classes_of_interest, index_inc=start)
        if labels is not None:
            z_block_labels = read_volume(labels, type=type, start=start, stop=stop)
            js_cum, ams_cum = _cumulate_validation_metrics(z_block_segmented, z_block_labels, js_cum, ams_cum,
                                                           classes_of_interest)

    else:

        # segment the first z-block
        z_block = read_volume(data, type=type, start=start, stop=start + zb)
        z_block_segmented = _segment_z_block(z_block, net, (zb, yb, xb), (zo, yo, xo), input_shape,
                                             in_channels=in_channels,
                                             batch_size=batch_size, step_size=step_size, train=train,
                                             track_progress=track_progress, device=device, orientations=orientations,
                                             normalization=normalization, bar=bar)

        # process the remaining z-blocks
        j = 0
        for i in range(zb, z, zb):
            z_start = i - zo
            z_stop = np.minimum(i + zb, z)
            z_block_prev = z_block_segmented
            z_block = read_volume(data, type=type, start=start + z_start, stop=start + z_stop)
            # overlapping region is in slices zb-zo through zb in z_block_prev and 0 through zo in z_block

            # segment z_block
            z_block_segmented = _segment_z_block(z_block, net, (zb, yb, xb), (zo, yo, xo), input_shape,
                                                 in_channels=in_channels, batch_size=batch_size, step_size=step_size,
                                                 train=train, track_progress=track_progress, device=device,
                                                 orientations=orientations, normalization=normalization, bar=bar)

            # merge overlapping regions in two z_blocks
            w = np.arange(1, zo + 1, 1) / zo
            w = np.ones((y, x))[np.newaxis, ...] * w[:, np.newaxis, np.newaxis]
            w_ = 1 - w
            for c in range(net.out_channels):
                z_block_prev[c, -zo:] = w * z_block_prev[c, -zo:] + w_ * z_block_segmented[c, :zo]

            # z_block_prev is completely segmented, ready to write out and validate
            if write_dir is not None:
                if i == zb:
                    _write_segmentation(z_block_prev, write_dir, classes_of_interest=classes_of_interest,
                                        index_inc=start)
                else:
                    _write_segmentation(z_block_prev[:, zo:], write_dir, classes_of_interest=classes_of_interest,
                                        index_inc=start + j * zb)
            if labels is not None:
                if i == zb:
                    z_block_labels = read_volume(labels, type=type, start=start, stop=start + zb)
                    js_cum, ams_cum = _cumulate_validation_metrics(z_block_prev, z_block_labels, js_cum, ams_cum,
                                                                   classes_of_interest)
                else:
                    z_block_labels = read_volume(labels, type=type, start=start + j * zb, stop=start + (j + 1) * zb)
                    js_cum, ams_cum = _cumulate_validation_metrics(z_block_prev[:, zo:], z_block_labels, js_cum,
                                                                   ams_cum, classes_of_interest)

            j += 1

        # write out and validate last block
        if write_dir is not None:
            _write_segmentation(z_block_segmented[:, zo:], write_dir, classes_of_interest=classes_of_interest,
                                index_inc=start + j * zb)
        if labels is not None:
            z_block_labels = read_volume(labels, type=type, start=start + j * zb, stop=start + z)
            js_cum, ams_cum = _cumulate_validation_metrics(z_block_segmented[:, zo:], z_block_labels, js_cum, ams_cum,
                                                           classes_of_interest)

    # compute validation metrics
    js, ams = _compute_validation_metrics(js_cum, ams_cum, classes_of_interest)

    # report validation metrics if necessary
    if track_progress:
        print()
    if report:
        _report_validation(js, ams, classes_of_interest)

    # save the results if necessary
    if val_file is not None:
        _save_validation(js, ams, val_file)

    return js, ams


def segment_ram(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False, track_progress=False,
                device=0, orientations=(0,), normalization='unit', write_dir=None, labels=None, val_file=None,
                classes_of_interest=(0, 1), hausdorff=False, report=True):
    """
    Segment a 3D image using a specific network

    :param data: 3D array (Z, Y, X) representing the 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of
    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientations: list of orientations to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :param write_dir: destination for the segmentation, set this equal to None, in which case only validation will be performed
    :param labels: directory that contains a 3D array (Z, Y, X) representing the 3D labels
    :param val_file: destination of the validation file (only required if labels is not None)
    :param classes_of_interest: index of the label of interest (required if labels is not None or write_dir is not None)
    :param hausdorff: compute hausdorff or not (only required if labels is not None)
    :param report: print the validation report or not (only required if labels is not None)
    :return: the segmented image
    """

    # compute segmentation for each orientation and average results
    segmentation = segment(data, net, input_shape, in_channels=in_channels, batch_size=batch_size, step_size=step_size,
                           train=train, track_progress=track_progress, device=device, orientations=orientations,
                           normalization=normalization)

    # write out the segmentation if necessary
    if write_dir is not None:
        _write_segmentation(segmentation, write_dir, classes_of_interest=classes_of_interest)

    # validate the segmentation if necessary
    if labels is not None:
        js, ams = _validate_ram(segmentation, labels, classes_of_interest=classes_of_interest, hausdorff=hausdorff,
                                report=report)
        # save the results if necessary
        if val_file is not None:
            _save_validation(js, ams, val_file)
    else:
        js, ams = None, None

    return segmentation, js, ams


def segment(data, net, input_shape, in_channels=1, batch_size=1, step_size=None, train=False, track_progress=False,
            device=0, orientations=(0,), normalization='unit'):
    """
    Segment a 3D image using a specific network.
    This function assumes there is plenty of RAM memory available to store the volume, its segmentation and all
    remaining temporary variables. For large volumes this can become quite memory intensive. In that case, please
    consider the segment_read function.

    :param data: 3D array (Z, Y, X) representing the 3D image
    :param net: image-to-image segmentation network
    :param input_shape: size of
    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
    :param batch_size: batch size for processing
    :param step_size: step size of the sliding window
    :param train: evaluate the network in training mode
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientations: list of orientations to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :return: the segmented image
    """

    # compute segmentation for each orientation and average results
    segmentation = np.zeros((net.out_channels, *data.shape))
    for orientation in orientations:
        segmentation += segment_multichannel(data[np.newaxis, ...], net, input_shape, in_channels=in_channels,
                                             batch_size=batch_size, step_size=step_size, train=train,
                                             track_progress=track_progress, device=device, orientation=orientation,
                                             normalization=normalization)
    segmentation = segmentation / len(orientations)

    return segmentation


def validate(net, data, labels, input_size, in_channels=1, classes_of_interest=(0, 1), batch_size=1, write_dir=None,
             val_file=None, track_progress=False, device=0, orientations=(0,), normalization='unit', hausdorff=False,
             report=True):
    """
    Validate a network on a dataset and its labels

    :param net: image-to-image segmentation network
    :param data: 3D array (Z, Y, X) representing the 3D image, or a directory that specifies to it (if RAM is limited)
    :param labels: 3D array (Z, Y, X) representing the 3D labels, or a directory that specifies to it (if RAM is limited)
    :param input_size: size of the inputs (either 2 or 3-tuple) for processing
    :param in_channels: Amount of subsequent slices that serve as input for the network (should be odd)
    :param classes_of_interest: index of the label of interest
    :param batch_size: batch size for processing
    :param write_dir: optionally, specify a directory to write the output
    :param val_file: optionally, specify a file to write the validation results
    :param track_progress: optionally, for tracking progress with progress bar
    :param device: GPU device where the computations should occur
    :param orientations: list of orientations to perform segmentation: 0-Z, 1-Y, 2-X (only for 2D based segmentation)
    :param normalization: type of data normalization (unit, z or minmax)
    :param hausdorff: compute hausdorff or not
    :param report: get a report of the validation results or not
    :return: validation results, i.e. accuracy, precision, recall, f-score, jaccard and dice score
    """

    if data.__class__ == str:
        js, ams = segment_read(data, net, input_size, write_dir=write_dir, in_channels=in_channels,
                               batch_size=batch_size, track_progress=track_progress, device=device,
                               orientations=orientations, normalization=normalization, labels=labels, val_file=val_file,
                               classes_of_interest=classes_of_interest, report=report)
    else:
        segmentation, js, ams = segment_ram(data, net, input_size, write_dir=write_dir, in_channels=in_channels,
                                            batch_size=batch_size, track_progress=track_progress, device=device,
                                            orientations=orientations, normalization=normalization, labels=labels,
                                            val_file=val_file, classes_of_interest=classes_of_interest,
                                            hausdorff=hausdorff, report=report)

    return js, ams




import warnings

from neuralnets.util.tools import sample_unlabeled_input, sample_labeled_input, normalize
from neuralnets.util.augmentation import split_segmentation_transforms
from neuralnets.data.base import *


MAX_SAMPLING_ATTEMPTS = 20


def _orient(data, orientation=0):
    """
    This function essentially places the desired orientation axis to that of the original Z-axis
    For example:
          (Z, Y, X) -> (Y, Z, X) for orientation=1
          (Z, Y, X) -> (X, Y, Z) for orientation=2
    Note that applying this function twice corresponds to the identity transform

    :param data: assumed to be of shape (Z, Y, X)
    :param orientation: 0, 1 or 2 (respectively for Z, Y or X axis)
    :return: reoriented data sample
    """
    if orientation == 1:
        return np.transpose(data, axes=(1, 0, 2))
    elif orientation == 2:
        return np.transpose(data, axes=(2, 1, 0))
    else:
        return data


def _validate_shape(input_shape, data_shape, orientation=0, in_channels=1, levels=4):
    """
    Validates an input for propagation through a U-Net by taking into account the following:
        - Sampling along different orientations
        - Sampling multiple adjacent slices as channels
        - Maximum size that can be sampled from the data

    :param input_shape: original shape of the sample (Z, Y, X)
    :param data_shape: shape of the data to sample from (Z, Y, X)
    :param orientation: orientation to sample along
    :param in_channels: sample multiple adjacent slices as channels
    :param levels: amount of pooling layers in the network
    :return: the validated input shape
    """

    # make sure input shape can be edited
    input_shape = list(input_shape)

    # sample adjacent slices if necessary
    is2d = input_shape[0] == 1
    if is2d and in_channels > 1:
        input_shape[0] = in_channels

    # transform the data shape and input shape according to the orientation
    if orientation == 1:  # (Z, Y, X) -> (Y, Z, X)
        input_shape = [input_shape[1], input_shape[0], input_shape[2]]
    elif orientation == 2:  # (Z, Y, X) -> (X, Y, Z)
        input_shape = [input_shape[2], input_shape[1], input_shape[0]]

    # make sure the input shape fits in the data shape: i.e. find largest k such that n of the form n=k*2**levels
    for d in range(3):
        if not (is2d and d == orientation) and input_shape[d] > data_shape[d]:
            # 2D case: X, Y - 3D case: X, Y, Z
            # note we assume that the data has a least in_channels elements in each dimension
            input_shape[d] = int((data_shape[d] // (2 ** levels)) * (2 ** levels))

    # and return as a tuple
    return tuple(input_shape)


def _map_cois(y, coi):
    """
    Maps the classes of interest to consecutive label indices

    :param y: labels
    :param coi: classes of interest
    :return: reindexed labels
    """
    coi_ = list(coi)
    coi_.sort()
    for i, c in enumerate(coi_):
        y[y == c] = i

    return y


class LabeledStandardDataset(StandardDataset):
    """
    Strongly labeled dataset of N 2D images and pixel-wise labels

    :param data_path: path to the dataset
    :param label_path: path to the labels
    :param optional scaling: tuple used for rescaling the data, or None
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional data_dtype: type of the data (typically uint8)
    :param optional label_dtype: type of the labels (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    """

    def __init__(self, data_path, label_path, scaling=None, type='tif3d', data_dtype='uint8', label_dtype='uint8',
                 coi=(0, 1), norm_type='unit'):
        super().__init__(data_path, scaling=scaling, type=type, dtype=data_dtype, norm_type=norm_type)

        self.label_path = label_path
        self.coi = coi

        # load labels
        self.labels = read_volume(label_path, type=type, dtype=label_dtype)

        # rescale the dataset if necessary
        if scaling is not None:
            target_size = np.asarray(np.multiply(self.labels.shape, scaling), dtype=int)
            self.labels = \
                F.interpolate(torch.Tensor(self.labels[np.newaxis, np.newaxis, ...]), size=tuple(target_size),
                              mode='area')[0, 0, ...].numpy()

        self.mu, self.std = self._get_stats()

    def __getitem__(self, i):

        # get random sample
        input = normalize(self.data[i], type=self.norm_type)
        target = self.labels[i]

        if input.shape[0] > 1:
            # add channel axis if the data is 3D
            input, target = input[np.newaxis, ...], target[np.newaxis, ...]

        if len(np.intersect1d(np.unique(target),
                              self.coi)) == 0:  # make sure we have at least one labeled pixel in the sample, otherwise processing is useless
            return self.__getitem__(i)
        else:
            return input, target


class UnlabeledStandardDataset(StandardDataset):
    """
    Unlabeled dataset of N 2D images

    :param data_path: path to the dataset
    :param optional scaling: tuple used for rescaling the data, or None
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional dtype: type of the data (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    """

    def __init__(self, data_path, scaling=None, type='tif3d', dtype='uint8', norm_type='unit'):
        super().__init__(data_path, scaling=scaling, type=type, dtype=dtype, norm_type=norm_type)

        self.mu, self.std = self._get_stats()

    def __getitem__(self, i):

        # get random sample
        input = normalize(self.data[i], type=self.norm_type)

        if input.shape[0] > 1:
            # add channel axis if the data is 3D
            return input[np.newaxis, ...]
        else:
            return input


class LabeledVolumeDataset(VolumeDataset):
    """
    Dataset for pixel-wise labeled volumes

    :param data: input data, possible formats
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param labels: path to the labels or a 3D volume that has already been loaded, possible formats:
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param input_shape: 3-tuple that specifies the input shape for sampling
    :param optional scaling: 3-tuple used for rescaling the data, or a list of 3-tuples in case of multiple datasets
    :param optional len_epoch: number of iterations for one epoch
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional coi: list or sequence of the classes of interest
    :param optional in_channels: amount of subsequent slices to be sampled (only for 2D sampling)
    :param optional orientations: list of orientations for sampling
    :param optional batch_size: size of the sampling batch
    :param optional data_dtype: type of the data (typically uint8)
    :param optional label_dtype: type of the labels (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    :param optional transform: augmenter object
    :param optional range_split: range of slices (start, stop) to select (normalized between 0 and 1), or a list of
                                 2-tuples in case of multiple datasets
    :param optional range_dir: orientation of the slicing or a list of orientations in case of multiple datasets
    :param optional resolution: list of 3-tuples specifying the pixel resolution of the data
    :param optional match_resolution_to: match the resolution of all data to a specific dataset
    :param optional sampling_type: type of sampling in case of multiple datasets
            - joint: the dataset will generate random samples in each dataset and return all of them
            - single: the dataset will generate a random sample from a randomly selected dataset and return that
    """

    def __init__(self, data, labels, input_shape=None, scaling=None, len_epoch=None, type='tif3d', coi=(0, 1),
                 in_channels=1, orientations=(0,), batch_size=1, data_dtype='uint8', label_dtype='uint8',
                 norm_type='unit', transform=None, range_split=None, range_dir=None, resolution=None,
                 match_resolution_to=None, sampling_type='joint'):
        super().__init__(data, input_shape, scaling=scaling, len_epoch=len_epoch, type=type,
                         in_channels=in_channels, orientations=orientations, batch_size=batch_size, dtype=data_dtype,
                         norm_type=norm_type, range_split=range_split, range_dir=range_dir, resolution=resolution,
                         match_resolution_to=match_resolution_to, sampling_type=sampling_type)

        if isinstance(labels, str) or isinstance(labels, np.ndarray):
            self.labels = [load_data(labels, data_type=type, dtype=label_dtype)]
        elif isinstance(labels, list) or isinstance(labels, tuple):  # list of data
            self.labels = []
            for labels_i in labels:
                self.labels.append(load_data(labels_i, data_type=type, dtype=label_dtype))
        else:
            raise ValueError('LabeledVolumeDataset requires labels in str, np.ndarray or list format')
        self.coi = coi
        self.transform = transform
        if transform is not None:
            self.shared_transform, self.x_transform, self.y_transform = split_segmentation_transforms(transform)

        # select a subset of slices of the data
        for i in range(len(self.labels)):
            if isinstance(range_split, list) and isinstance(range_dir, list):
                self.labels[i] = slice_subset(self.labels[i], range_split[i], range_dir[i])
            else:
                self.labels[i] = slice_subset(self.labels[i], range_split, range_dir)

        # rescale the dataset if necessary
        for i in range(len(self.labels)):
            if (self.resolution is not None and self.match_resolution_to is not None) or self.scaling is not None:
                if self.resolution is not None and self.match_resolution_to is not None:
                    scale_factor = np.divide(self.labels[self.match_resolution_to].shape, self.labels[i].shape)
                else:
                    if isinstance(self.scaling, list):
                        scale_factor = self.scaling[i]
                    else:
                        scale_factor = self.scaling
                labels_it = torch.Tensor(self.labels[i]).unsqueeze(0).unsqueeze(0)
                self.labels[i] = F.interpolate(labels_it, scale_factor=scale_factor, mode='bilinear')[0, 0, ...].numpy()

        # relabel classes of interest
        self.labels = [_map_cois(l, self.coi) for l in self.labels]

    def _sample(self, data_index):

        # get shape of sample
        input_shape = _validate_shape(self.input_shape, self.data[data_index].shape, in_channels=self.in_channels,
                                      orientation=self.orientation)

        # get random sample
        x, y = sample_labeled_input(self.data[data_index], self.labels[data_index], input_shape)
        x = normalize(x, type=self.norm_type)
        y = y.astype(float)

        # reorient sample
        x = _orient(x, orientation=self.orientation)
        y = _orient(y, orientation=self.orientation)

        # add channel axis if the data is 3D
        if self.input_shape[0] > 1:
            x, y = x[np.newaxis, ...], y[np.newaxis, ...]

        # select middle slice if multiple consecutive slices
        if self.in_channels > 1:
            c = self.in_channels // 2
            y = y[c:c + 1]

        # augment sample
        if self.transform is not None:
            data = self.shared_transform(np.concatenate((x, y), axis=0))
            p = x.shape[0]
            x = self.x_transform(data[:p])
            y = self.y_transform(data[p:])

        # transform to tensors
        x = torch.from_numpy(x).float()
        y = torch.from_numpy(y).long()

        return x, y

    def __getitem__(self, i, attempt=0):

        # reorient when we start a new batch
        if i % self.batch_size == 0:
            self._select_orientation()

        if self.sampling_type == 'single':

            # randomly select a dataset
            r = np.random.randint(len(self.data))

            # select a sample from dataset r
            x, y = self._sample(r)

            # make sure we have at least one labeled pixel in the sample, otherwise processing is useless
            if len(np.intersect1d(torch.unique(y).numpy(), self.coi)) == 0 and not self.warned:
                if attempt < MAX_SAMPLING_ATTEMPTS:
                    x, y = self.__getitem__(i, attempt=attempt + 1)
                else:
                    warnings.warn("No labeled pixels found after %d sampling attempts! " % attempt)
                    self.warned = True

        else:  # joint sampling

            xs = []
            ys = []

            for r in range(len(self.data)):

                # select a sample from dataset r
                x, y = self._sample(r)

                xs.append(x)
                ys.append(y)

            if len(self.data) == 1:
                x = xs[0]
                y = ys[0]
            else:
                x = xs
                y = ys

            # make sure we have at least one labeled pixel in the sample, otherwise processing is useless
            if np.sum([len(np.intersect1d(torch.unique(y_).numpy(), self.coi)) for y_ in ys]) == 0 and not self.warned:
                if attempt < MAX_SAMPLING_ATTEMPTS:
                    x, y = self.__getitem__(i, attempt=attempt + 1)
                else:
                    warnings.warn("No labeled pixels found after %d sampling attempts! " % attempt)
                    self.warned = True

        # return sample
        return x, y


class UnlabeledVolumeDataset(VolumeDataset):
    """
    Dataset for unlabeled volumes

    :param data: input data, possible formats
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param input_shape: 3-tuple that specifies the input shape for sampling
    :param optional scaling: 3-tuple used for rescaling the data, or a list of 3-tuples in case of multiple datasets
    :param optional len_epoch: number of iterations for one epoch
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional in_channels: amount of subsequent slices to be sampled (only for 2D sampling)
    :param optional orientations: list of orientations for sampling
    :param optional batch_size: size of the sampling batch
    :param optional dtype: type of the data (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    :param optional transform: augmenter object
    :param optional range_split: range of slices (start, stop) to select (normalized between 0 and 1), or a list of
                                 2-tuples in case of multiple datasets
    :param optional range_dir: orientation of the slicing or a list of orientations in case of multiple datasets
    :param optional resolution: list of 3-tuples specifying the pixel resolution of the data
    :param optional match_resolution_to: match the resolution of all data to a specific dataset
    :param optional sampling_type: type of sampling in case of multiple datasets
            - joint: the dataset will generate random samples in each dataset and return all of them
            - single: the dataset will generate a random sample from a randomly selected dataset and return that
    """

    def __init__(self, data, input_shape=None, scaling=None, len_epoch=None, type='tif3d', in_channels=1,
                 orientations=(0,), batch_size=1, dtype='uint8', norm_type='unit', transform=None, range_split=None,
                 range_dir=None, resolution=None, match_resolution_to=None, sampling_type='joint'):
        super().__init__(data, input_shape, scaling=scaling, len_epoch=len_epoch, type=type,
                         in_channels=in_channels, orientations=orientations, batch_size=batch_size, dtype=dtype,
                         norm_type=norm_type, range_split=range_split, range_dir=range_dir, resolution=resolution,
                         match_resolution_to=match_resolution_to, sampling_type=sampling_type)

        self.transform = transform

    def _sample(self, data_index):

        # get shape of sample
        input_shape = _validate_shape(self.input_shape, self.data[data_index].shape, in_channels=self.in_channels,
                                      orientation=self.orientation)

        # get random sample
        x = sample_unlabeled_input(self.data[data_index], input_shape)
        x = normalize(x, type=self.norm_type)

        # reorient sample
        x = _orient(x, orientation=self.orientation)

        # add channel axis if the data is 3D
        if self.input_shape[0] > 1:
            x = x[np.newaxis, ...]

        # augment sample
        if self.transform is not None:
            x = self.transform(x)

        # transform to tensors
        x = torch.from_numpy(x).float()

        return x

    def __getitem__(self, i):

        # reorient when we start a new batch
        if i % self.batch_size == 0:
            self._select_orientation()

        if self.sampling_type == 'single':

            # randomly select a dataset
            r = np.random.randint(len(self.data))

            # select a sample from dataset r
            x = self._sample(r)

        else:  # joint sampling

            xs = []

            for r in range(len(self.data)):

                # select a sample from dataset r
                x = self._sample(r)

                xs.append(x)

            if len(self.data) == 1:
                x = xs[0]
            else:
                x = xs

        # return sample
        return x


class LabeledSlidingWindowDataset(SlidingWindowDataset):
    """
    Dataset for pixel-wise labeled volumes with a sliding window

    :param data: input data, possible formats
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param labels: path to the labels or a 3D volume that has already been loaded, possible formats:
            - path to the dataset (string)
            - preloaded 3D volume (numpy array)
            - list of paths to multiple datasets (list of strings)
            - list of preloaded 3D volumes (list of numpy arrays)
    :param input_shape: 3-tuple that specifies the input shape for sampling
    :param optional scaling: 3-tuple used for rescaling the data, or a list of 3-tuples in case of multiple datasets
    :param optional type: type of the volume file (tif2d, tif3d, tifseq, hdf5, png or pngseq)
    :param optional in_channels: amount of subsequent slices to be sampled (only for 2D sampling)
    :param optional orientations: list of orientations for sampling
    :param optional coi: list or sequence of the classes of interest
    :param optional batch_size: size of the sampling batch
    :param optional data_dtype: type of the data (typically uint8)
    :param optional label_dtype: type of the labels (typically uint8)
    :param optional norm_type: type of the normalization (unit, z or minmax)
    :param optional transform: augmenter object
    :param optional range_split: range of slices (start, stop) to select (normalized between 0 and 1), or a list of
                                 2-tuples in case of multiple datasets
    :param optional range_dir: orientation of the slicing or a list of orientations in case of multiple datasets
    :param optional resolution: list of 3-tuples specifying the pixel resolution of the data
    :param optional match_resolution_to: match the resolution of all data to a specific dataset
    """

    def __init__(self, data, labels, input_shape=None, scaling=None, type='tif3d', in_channels=1, orientations=(0,),
                 coi=(0, 1), batch_size=1, data_dtype='uint8', label_dtype='uint8', norm_type='unit', transform=None,
                 range_split=None, range_dir=None, resolution=None, match_resolution_to=None):
        super().__init__(data, input_shape, scaling=scaling, type=type, in_channels=in_channels,
                         orientations=orientations, batch_size=batch_size, dtype=data_dtype, norm_type=norm_type,
                         range_split=range_split, range_dir=range_dir, resolution=resolution,
                         match_resolution_to=match_resolution_to)

        if isinstance(labels, str) or isinstance(labels, np.ndarray):
            self.labels = [load_data(labels, data_type=type, dtype=label_dtype)]
        elif isinstance(labels, list) or isinstance(labels, tuple):  # list of data
            self.labels = []
            for labels_i in labels:
                self.labels.append(load_data(labels_i, data_type=type, dtype=label_dtype))
        else:
            raise ValueError('LabeledSlidingWindowDataset requires labels in str, np.ndarray or list format')
        self.coi = coi
        self.transform = transform
        if transform is not None:
            self.shared_transform, self.x_transform, self.y_transform = split_segmentation_transforms(transform)

        # select a subset of slices of the data
        for i in range(len(self.labels)):
            if isinstance(range_split, list) and isinstance(range_dir, list):
                self.labels[i] = slice_subset(self.labels[i], range_split[i], range_dir[i])
            else:
                self.labels[i] = slice_subset(self.labels[i], range_split, range_dir)

        # rescale the dataset if necessary
        for i in range(len(self.labels)):
            if (self.resolution is not None and self.match_resolution_to is not None) or self.scaling is not None:
                if self.resolution is not None and self.match_resolution_to is not None:
                    scale_factor = np.divide(self.labels[self.match_resolution_to].shape, self.labels[i].shape)
                else:
                    if isinstance(self.scaling, list):
                        scale_factor = self.scaling[i]
                    else:
                        scale_factor = self.scaling
                labels_it = torch.Tensor(self.labels[i]).unsqueeze(0).unsqueeze(0)
                self.labels[i] = F.interpolate(labels_it, scale_factor=scale_factor, mode='bilinear')[0, 0, ...].numpy()

        # pad data so that the dimensions are a multiple of the inputs shapes
        self.labels = [pad2multiple(l, input_shape, value=255) for l in self.labels]

        # pad data so that additional channels can be sampled
        self.labels = [pad_channels(l, in_channels=in_channels, orientations=self.orientations) for l in self.labels]

        # relabel classes of interest
        self.labels = [_map_cois(l, self.coi) for l in self.labels]

    def __getitem__(self, i):

        # find dataset index
        r = 0
        szs = self.n_samples_dim.prod(axis=1)
        while szs[:r + 1].sum() <= i:
            r += 1

        # get spatial location
        j = i - szs[:r].sum()
        iz = j // (self.n_samples_dim[r, 1] * self.n_samples_dim[r, 2])
        iy = (j - iz * self.n_samples_dim[r, 1] * self.n_samples_dim[r, 2]) // self.n_samples_dim[r, 2]
        ix = j - iz * self.n_samples_dim[r, 1] * self.n_samples_dim[r, 2] - iy * self.n_samples_dim[r, 2]
        pz = self.input_shape[0] * iz
        py = self.input_shape[1] * iy
        px = self.input_shape[2] * ix

        # get shape of sample
        input_shape = _validate_shape(self.input_shape, self.data[r].shape, in_channels=self.in_channels)

        # get sample
        x, y = sample_labeled_input(self.data[r], self.labels[r], input_shape, zyx=(pz, py, px))
        x = normalize(x, type=self.norm_type)
        y = y.astype(float)

        # add channel axis if the data is 3D
        if self.input_shape[0] > 1:
            x, y = x[np.newaxis, ...], y[np.newaxis, ...]

        # select middle slice if multiple consecutive slices
        if self.in_channels > 1:
            c = self.in_channels // 2
            y = y[c:c + 1]

        # augment sample
        if self.transform is not None:
            data = self.shared_transform(np.concatenate((x, y), axis=0))
            p = x.shape[0]
            x = self.x_transform(data[:p])
            y = self.y_transform(data[p:])

        # transform to tensors
        x = torch.from_numpy(x).float()
        y = torch.from_numpy(y).long()

        # make sure we have at least one labeled pixel in the sample, otherwise processing is useless
        if len(np.intersect1d(torch.unique(y).numpy(), np.arange(len(self.coi)))) == 0 and not self.warned:
            warnings.warn("No labeled pixels found! ")
            self.warned = True

        # return sample
        return x, y



import torch


def iou(y_true, y_pred, w=None):
    """
    Intersection-over-union metric

    :param y_true: (N1, N2, ...) tensor of the true labels
    :param y_pred: (N1, N2, ...) tensor of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking tensor
    :return: the IoU
    """

    # threshold
    y_pred = y_pred > 0.5

    # mask if necessary
    if w is not None:
        y_pred = y_pred[w]
        y_true = y_true[w]

    # compute jaccard score
    intersection = torch.sum(y_pred * y_true)
    union = torch.sum(y_pred) + torch.sum(y_true) - intersection

    return (intersection + 1) / (union + 1)


def dice(y_true, y_pred, w=None):
    """
    Dice coefficient between two segmentations

    :param y_true: (N1, N2, ...) tensor of the true labels
    :param y_pred: (N1, N2, ...) tensor of the predictions (either probs or binary)
    :param w: (N1, N2, ...) masking tensor
    :return: the Dice coefficient
    """

    j = iou(y_true, y_pred, w=w)

    return 2 * j / (1 + j)



"""
    This is a script that illustrates grid search cross validation to optimize parameters for a 2D U-Net
"""

"""
    Necessary libraries
"""
import argparse
import yaml
import os

from neuralnets.data.datasets import LabeledVolumeDataset
from neuralnets.util.augmentation import *
from neuralnets.util.io import print_frm, save
from neuralnets.util.tools import set_seed, parse_params, log_hparams
from neuralnets.cross_validation.base import UNet2DClassifier, parse_search_grid

from multiprocessing import freeze_support
from sklearn.model_selection import GridSearchCV


if __name__ == '__main__':
    freeze_support()

    """
        Parse all the arguments
    """
    print_frm('Parsing arguments')
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", "-c", help="Path to the configuration file", type=str, default='unet_2d.yaml')
    args = parser.parse_args()
    with open(args.config) as file:
        params = parse_params(yaml.load(file, Loader=yaml.FullLoader))

    """
    Fix seed (for reproducibility)
    """
    set_seed(params['seed'])

    """
        Load the data
    """
    print_frm('Loading data')
    input_shape = (1, *(params['input_size']))
    split = params['train_val_test_split']
    transform = Compose([Rotate90(), Flip(prob=0.5, dim=0), Flip(prob=0.5, dim=1), ContrastAdjust(adj=0.1),
                         RandomDeformation(), AddNoise(sigma_max=0.05), CleanDeformedLabels(params['coi'])])
    train = LabeledVolumeDataset(params['data'], params['labels'], input_shape=input_shape, type=params['type'],
                                 batch_size=params['train_batch_size'], transform=transform, range_split=(0, split[0]),
                                 range_dir=params['split_orientation'])
    X_train = train.data
    y_train = train.labels

    """
        Build the network
    """
    print_frm('Building the network')
    clf = UNet2DClassifier(epochs=params['epochs'], gpus=params['gpus'], accelerator=params['accelerator'],
                           log_dir=params['log_dir'], log_freq=params['log_freq'],
                           log_refresh_rate=params['log_refresh_rate'], train_batch_size=params['train_batch_size'],
                           test_batch_size=params['test_batch_size'], num_workers=params['num_workers'],
                           device=params['gpus'][0], transform=transform, feature_maps=params['fm'],
                           levels=params['levels'], dropout=params['dropout'], norm=params['norm'],
                           activation=params['activation'], coi=params['coi'], loss_fn=params['loss'])

    """
        Perform cross validation grid search
    """
    print_frm('Starting grid search cross validation')
    search_grid = parse_search_grid(params['search_grid'])
    gs = GridSearchCV(clf, search_grid, cv=params['folds'], verbose=4)
    gs.fit(X_train, y_train)

    """
        Save and report results
    """
    save(gs.cv_results_, params['results_file'])
    hparams_dir = os.path.join(params['log_dir'], 'hparams')
    log_hparams(gs, log_dir=hparams_dir)
    print_frm(gs.best_params_)
    print_frm('Best mIoU: %.6f' % gs.best_score_)

    print_frm('Finished!')
